from django.contrib import admin
from enumfields.admin import EnumFieldListFilter

from arkindex.documents.models import (
    AllowedMetaData,
    Classification,
    Corpus,
    CorpusExport,
    Element,
    ElementType,
    Entity,
    EntityType,
    MetaData,
    MLClass,
    Transcription,
)


class ElementTypeInline(admin.TabularInline):
    model = ElementType
    fields = ("slug", "display_name", "folder", "indexable")
    readonly_fields = ("slug", "display_name", "folder")

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Corpus)
class CorpusAdmin(admin.ModelAdmin):
    fields = ("id", "name", "description", "top_level_type", "public", "indexable", "maximum_task_ttl", "created")
    readonly_fields = ("id", "created")
    list_display = ("id", "name", "public", "top_level_type", "created")
    search_fields = ("name", )
    inlines = (ElementTypeInline, )
    ordering = ("-created", )

    def has_delete_permission(self, request, obj=None):
        # Require everyone to use the asynchronous corpus deletion
        return False

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("top_level_type")

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        if obj:
            # Limit top level type queryset to the types of this corpus
            form.base_fields["top_level_type"].queryset = ElementType.objects.filter(corpus=obj)
        return form


@admin.register(CorpusExport)
class CorpusExportAdmin(admin.ModelAdmin):
    list_display = ("id", "corpus", "user", "state")
    ordering = ("-created", )


class ClassificationInline(admin.TabularInline):
    model = Classification
    readonly_fields = ("confidence", "high_confidence", )
    raw_id_fields = ("worker_version", "worker_run", "moderator", "ml_class")


@admin.register(AllowedMetaData)
class AllowedMetaDataAdmin(admin.ModelAdmin):
    list_display = ("id", "corpus", "type", "name")
    readonly_fields = ("id", )

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        form.base_fields["corpus"].queryset = Corpus.objects.order_by("name", "id")
        return form


@admin.register(MetaData)
class MetaDataAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "type", )
    list_filter = [("type", EnumFieldListFilter), ]
    readonly_fields = ("id", )
    raw_id_fields = ("element", "entity", "worker_version", "worker_run")


class MetaDataInline(admin.TabularInline):
    model = MetaData
    raw_id_fields = ("entity", "worker_version", "worker_run")


@admin.register(Element)
class ElementAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "type", "corpus", )
    list_filter = ["type__slug", "corpus"]
    fields = ("id", "type", "name", "image", "corpus", "confidence", "worker_run", )
    raw_id_fields = ("worker_run",)
    readonly_fields = ("id", "corpus", "image",)
    search_fields = ("name", )
    inlines = (MetaDataInline, ClassificationInline)

    # Disable element creation through the admin to prevent element_type conflicts
    def has_add_permission(self, request):
        return False

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "type":
            # Only display types available on the element's corpus
            # It would make no sense to pick a type from another corpus
            # Display all elements on creation, as we do not know yet the corpus
            element_id = request.resolver_match.kwargs.get("object_id")
            if element_id:
                element = self.get_object(request, element_id)
                kwargs["queryset"] = ElementType.objects.filter(corpus=element.corpus).order_by("display_name")

        return super().formfield_for_foreignkey(db_field, request, **kwargs)


@admin.register(Transcription)
class TranscriptionAdmin(admin.ModelAdmin):
    list_display = ("id", "text", "confidence", "orientation", "element", )
    fields = ("id", "text", "confidence", "orientation", "element", )
    readonly_fields = ("id", )
    raw_id_fields = ("element", )


@admin.register(MLClass)
class MLClassAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "corpus")
    list_filter = ("corpus",)
    search_fields = ("name",)
    fields = ("name", "corpus")

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        form.base_fields["corpus"].queryset = Corpus.objects.order_by("name", "id")
        return form


@admin.register(Entity)
class EntityAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "type")
    list_filter = ["corpus", "type"]
    readonly_fields = ("id", )
    raw_id_fields = ("worker_version", "worker_run", )
    search_fields = ("name", )


@admin.register(EntityType)
class EntityTypeAdmin(admin.ModelAdmin):
    list_display = ("id", "corpus", "name", "color")
    list_filter = ("corpus", )

    def get_readonly_fields(self, request, obj=None):
        # Make the corpus field read-only only for existing entity types.
        # Otherwise, new EntityTypes would be created with corpus=None
        if obj:
            return ("id", "corpus")
        return ("id", )

    def has_delete_permission(self, request, obj=None):
        # Require everyone to use the frontend or DestroyEntityType
        return False
