#!/usr/bin/env python3
from django.core.management.base import BaseCommand
from django.db import connections, transaction

from arkindex.project.argparse import CorpusArgument


class Command(BaseCommand):
    help = "Merge entities of the same name and type together in a corpus."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "-c",
            "--corpus",
            help="ID of the corpus to merge entities on.",
            type=CorpusArgument(),
            required=True,
        )

    @transaction.atomic
    def handle(self, corpus, **options):
        with connections["default"].cursor() as cursor:
            # Create a temporary table to map the duplicate entity IDs with the ones that will be kept.
            # We include a primary key and a unique constraint because those will both add some indexes,
            # which makes all joins faster on all subsequent queries — even if it is a temporary table.
            cursor.execute("DROP TABLE IF EXISTS duplicated_entities;")
            cursor.execute("""
            CREATE TEMPORARY TABLE duplicated_entities (
                remove_id UUID NOT NULL,
                keep_id UUID NOT NULL,
                PRIMARY KEY (remove_id, keep_id),
                UNIQUE (remove_id)
            );
            """)

            self.stdout.write(f"Looking for duplicate entities in {corpus.name}…")

            cursor.execute("""
            INSERT INTO duplicated_entities (remove_id, keep_id)
            SELECT remove_id, keep_id
            FROM (
                SELECT id AS remove_id, FIRST_VALUE(id) OVER (PARTITION BY name, type_id ORDER BY id) AS keep_id
                FROM documents_entity
                WHERE corpus_id = %s
            ) b WHERE remove_id <> keep_id;
            """, [corpus.id])

            self.stdout.write(f"{cursor.rowcount} duplicate entities found.")
            if cursor.rowcount < 1:
                # No duplicates, we're done here.
                cursor.execute("DROP TABLE duplicated_entities;")
                return

            self.stdout.write("Updating entity IDs on metadata…")
            cursor.execute("""
            UPDATE documents_metadata
            SET entity_id = keep_id
            FROM duplicated_entities
            WHERE entity_id = remove_id;
            """)
            self.stdout.write(f"Updated {cursor.rowcount} metadata.")

            self.stdout.write("Deleting duplicate TranscriptionEntities with worker runs…")
            cursor.execute("""
            DELETE FROM documents_transcriptionentity
            USING (
                SELECT id AS remove_id, FIRST_VALUE(id) OVER (
                    PARTITION BY
                        transcription_id,
                        duplicated_entities.keep_id,
                        "offset",
                        length,
                        worker_run_id
                    ORDER BY id
                ) AS keep_id
                FROM documents_transcriptionentity
                INNER JOIN duplicated_entities ON (entity_id IN (duplicated_entities.remove_id, duplicated_entities.keep_id))
                WHERE worker_run_id IS NOT NULL
            ) duplicates
            WHERE duplicates.remove_id <> duplicates.keep_id
            AND documents_transcriptionentity.id = duplicates.remove_id;
            """)
            self.stdout.write(f"Deleted {cursor.rowcount} TranscriptionEntities.")

            self.stdout.write("Deleting duplicate manual TranscriptionEntities…")
            cursor.execute("""
            DELETE FROM documents_transcriptionentity
            USING (
                SELECT id AS remove_id, FIRST_VALUE(id) OVER (
                    PARTITION BY
                        transcription_id,
                        duplicated_entities.keep_id,
                        "offset",
                        length
                    ORDER BY id
                ) AS keep_id
                FROM documents_transcriptionentity
                INNER JOIN duplicated_entities ON (entity_id IN (duplicated_entities.remove_id, duplicated_entities.keep_id))
                WHERE worker_run_id IS NULL
            ) duplicates
            WHERE duplicates.remove_id <> duplicates.keep_id
            AND documents_transcriptionentity.id = duplicates.remove_id;
            """)
            self.stdout.write(f"Deleted {cursor.rowcount} TranscriptionEntities.")

            self.stdout.write("Updating entity IDs on TranscriptionEntities…")
            cursor.execute("""
            UPDATE documents_transcriptionentity
            SET entity_id = keep_id
            FROM duplicated_entities
            WHERE entity_id = remove_id;
            """)
            self.stdout.write(f"Updated {cursor.rowcount} TranscriptionEntities.")

            self.stdout.write("Removing duplicate entities…")
            cursor.execute("""
            DELETE FROM documents_entity
            USING duplicated_entities
            WHERE id = remove_id;
            """)
            self.stdout.write(self.style.SUCCESS(f"Removed {cursor.rowcount} duplicate entities from corpus {corpus.name}."))

            cursor.execute("DROP TABLE duplicated_entities;")
