import logging

from django.apps import apps
from enumfields import Enum

logger = logging.getLogger(__name__)


DEFAULT_PROCESS_NAME = "Migration of ML results without worker runs"
ML_RESULT_MODELS = (
    ("Element", "corpus_id"),
    ("Transcription", "element__corpus_id"),
    ("Classification", "element__corpus_id"),
    ("MetaData", "element__corpus_id"),
    ("Entity", "corpus_id"),
    ("TranscriptionEntity", "entity__corpus_id")
)

def create_missing_worker_runs(*, get_model = apps.get_model):
    User = apps.get_model("users", "User")
    Process = apps.get_model("process", "Process")
    WorkerRun = apps.get_model("process", "WorkerRun")

    isolated_versions = set()
    for name, corpus_key in ML_RESULT_MODELS:
        model = apps.get_model("documents", name)
        isolated_versions.update(
            model.objects
            .filter(worker_run__isnull=True, worker_version__isnull=False)
            .values_list(corpus_key, "worker_version_id")
            .distinct()
        )
    if not isolated_versions:
        return

    logger.info(f"Found {len(isolated_versions)} isolated versions that will require to build a fake process with a worker run.")
    default_admin = User.objects.filter(is_admin=True).earliest("id")
    corpus_ids = set(corpus_id for corpus_id, _ in isolated_versions)

    class ProcessMode(Enum):
        Workers = "workers"

    # Initially fetch the existing processes, then create the missing ones
    processes = list(
        Process.objects.filter(
            name=DEFAULT_PROCESS_NAME,
            corpus__in=corpus_ids,
        )
        .distinct("corpus_id")
        .only("id", "corpus_id")
    )
    if processes:
        logger.info(f"Fetched {len(processes)} existing processes to build missing worker runs.")
    remaining_corpus_ids = corpus_ids - set(process.corpus_id for process in processes)
    if remaining_corpus_ids:
        new_processes = Process.objects.bulk_create(
            Process(
                corpus_id=corpus_id,
                creator=default_admin,
                mode=ProcessMode.Workers,
                name="Migration of ML results without worker runs"
            ) for corpus_id in remaining_corpus_ids
        )
        logger.info(f"Created {len(new_processes)} processes to hold new worker runs.")
        processes.extend(new_processes)

    # Map processes by corpus ID and create associated worker runs
    processes = {process.corpus_id: process for process in processes}

    # Initially fetch the existing worker runs, then create the missing ones
    worker_runs = list(
        WorkerRun.objects
        .filter(process__in=[p.id for p in processes.values()])
        .select_related("process")
        .only("process__corpus_id", "version_id")
    )
    if worker_runs:
        logger.info(f"Fetched {len(worker_runs)} existing worker runs.")
    remaining_isolated_versions = [
        (corpus_id, worker_version_id)
        for corpus_id, worker_version_id in isolated_versions
        if not any(
            run.process.corpus_id == corpus_id and run.version_id == worker_version_id
            for run in worker_runs
        )
    ]
    if remaining_isolated_versions:
        new_worker_runs = WorkerRun.objects.bulk_create(
            [
                WorkerRun(
                    process=processes[corpus_id],
                    version_id=worker_version_id,
                ) for corpus_id, worker_version_id in remaining_isolated_versions
            ]
        )
        logger.info(f"Created {len(new_worker_runs)} worker runs.")
        worker_runs.extend(new_worker_runs)

    # Map worker runs by corpus ID/version ID couples
    worker_runs = {
        (worker_run.process.corpus_id, worker_run.version_id): worker_run for worker_run in worker_runs
    }

    # Finally update all ML results by steps of corpus/versions
    for name, corpus_key in ML_RESULT_MODELS:
        model = apps.get_model("documents", name)
        qs = model.objects.filter(worker_run__isnull=True, worker_version__isnull=False)
        logger.info(f"Updating {qs.count()} {name}.")
        for i, (corpus_id, worker_version_id) in enumerate(isolated_versions):
            print(f"\rProgress: [{i}/{len(isolated_versions)}]", end="")
            instances = qs.filter(**{
                corpus_key: corpus_id,
                "worker_version_id": worker_version_id
            })
            instances.update(worker_run=worker_runs[(corpus_id, worker_version_id)])
        print()
