import logging
from typing import Literal, Optional
from uuid import UUID

from django.conf import settings
from django.core.mail import send_mail
from django.template.loader import render_to_string
from django_rq import job
from rq import Retry, get_current_job

from arkindex.documents.indexer import Indexer
from arkindex.documents.managers import ElementQuerySet
from arkindex.documents.models import (
    Classification,
    Corpus,
    Element,
    ElementPath,
    Entity,
    MetaData,
    Selection,
    Transcription,
    TranscriptionEntity,
)
from arkindex.ponos.models import Artifact, Task
from arkindex.process.models import Process, ProcessDatasetSet, ProcessElement, WorkerActivity, WorkerRun
from arkindex.training.models import DatasetElement, DatasetSet
from arkindex.users.models import User

logger = logging.getLogger(__name__)


@job("high", timeout=settings.RQ_TIMEOUTS["corpus_delete"], retry=Retry(max=3))
def corpus_delete(corpus_id: str) -> None:
    # Note that this can be None when the task is run outside of a RQ worker (e.g. unit test)
    rq_job = get_current_job()
    corpus = Corpus.objects.get(id=corpus_id)
    logger.info(f"Deleting {corpus!r}")

    # When a corpus has a top level type, deleting all the ElementTypes will cause an IntegrityError.
    if corpus.top_level_type_id is not None:
        corpus.top_level_type_id = None
        corpus.save(update_fields=["top_level_type_id"])

    # Set process.element to NULL in all processes, to avoid IntegrityErrors when deleting elements, as
    # elements are deleted before processes.
    Process.objects.filter(corpus_id=corpus_id).exclude(element_id=None).update(element_id=None)

    # Delete all related objects, bypassing RESTRICTs deletion rules
    # and Django's way of loading everything into memory before deleting.
    querysets = [
        corpus.allowed_metadatas.all(),
        corpus.worker_version_cache.all(),
        # Using two separate queries is a lot faster than using a OR for large corpora
        ProcessElement.objects.filter(process__corpus_id=corpus_id),
        ProcessElement.objects.filter(element__corpus_id=corpus_id),
        # Process-DataFile M2M with implicit model
        Process.files.through.objects.filter(process__corpus_id=corpus_id),
        Process.files.through.objects.filter(datafile__corpus_id=corpus_id),
        # Worker activities are deleted in two queries, as filtering using OR is slower
        WorkerActivity.objects.filter(element__corpus_id=corpus_id),
        WorkerActivity.objects.filter(process__corpus_id=corpus_id),
        corpus.files.all(),
        MetaData.objects.filter(element__corpus_id=corpus_id),
        TranscriptionEntity.objects.filter(entity__corpus_id=corpus_id),
        TranscriptionEntity.objects.filter(transcription__element__corpus_id=corpus_id),
        corpus.entities.all(),
        corpus.entity_types.all(),
        Classification.objects.filter(element__corpus_id=corpus_id),
        Transcription.objects.filter(element__corpus_id=corpus_id),
        ElementPath.objects.filter(element__corpus_id=corpus_id),
        Selection.objects.filter(element__corpus_id=corpus_id),
        corpus.memberships.all(),
        corpus.exports.all(),
        # Delete task artifacts
        Artifact.objects.filter(task__process__corpus_id=corpus_id),
        # ProcessDatasetSet M2M
        ProcessDatasetSet.objects.filter(set__dataset__corpus_id=corpus_id),
        ProcessDatasetSet.objects.filter(process__corpus_id=corpus_id),
        DatasetElement.objects.filter(set__dataset__corpus_id=corpus_id),
        DatasetSet.objects.filter(dataset__corpus_id=corpus_id),
        corpus.datasets.all(),
        # Delete the hidden M2M task parents table
        Task.parents.through.objects.filter(from_task__process__corpus_id=corpus_id),
        Task.objects.filter(process__corpus_id=corpus_id),
        corpus.elements.all(),
        WorkerRun.objects.filter(process__corpus_id=corpus_id),
        corpus.processes.all(),
        corpus.types.all(),
        corpus.ml_classes.all(),
        Corpus.objects.filter(id=corpus_id),
    ]

    for i, queryset in enumerate(querysets):
        if rq_job:
            rq_job.set_progress(i / len(querysets))
            rq_job.set_description(f"Deleting {queryset.model.__name__} on corpus {corpus.name}")
        deleted_count = queryset._raw_delete(using="default")
        logger.info(f"Deleted {deleted_count} {queryset.model.__name__}")

    logger.info(f"Deleted corpus {corpus_id}")


@job("high", timeout=settings.RQ_TIMEOUTS["element_trash"])
def element_delete(element: Element, delete_children: bool) -> None:
    """
    Wrapper around the element_trash task that removes the element from its children's paths
    when not deleting recursively.
    """
    if not delete_children:
        element.remove_children()

    element_trash(
        Element.objects.filter(id=element.id),
        delete_children=delete_children,
    )


@job("high", timeout=settings.RQ_TIMEOUTS["element_trash"])
def element_trash(queryset: ElementQuerySet, delete_children: bool) -> None:
    queryset.trash(delete_children=delete_children)


@job("high", timeout=settings.RQ_TIMEOUTS["worker_results_delete"])
def selection_worker_results_delete(
    corpus_id: str,
    worker_run_id: Optional[str] = None,
    model_version_id: Optional[str] = None,
    configuration_id: Optional[str | Literal[False]] = None,
    version_id: Optional[str] = None,
) -> None:
    """
    Delete all Worker Results produced by any WorkerVersion or a specific one
    on a user's selected elements, including their children.
    """
    rq_job = get_current_job()
    assert rq_job is not None, "This task can only be run in a RQ job context."
    assert rq_job.user_id is not None, "This task requires a user ID to be defined on the RQ job."

    queryset = Selection.objects.filter(user_id=rq_job.user_id, element__corpus_id=corpus_id)
    total = queryset.count()
    for i, element_id in enumerate(queryset.values_list("element_id", flat=True).iterator()):
        rq_job.set_progress(i / total)
        worker_results_delete(
            corpus_id=corpus_id,
            element_id=element_id,
            worker_run_id=worker_run_id,
            version_id=version_id,
            model_version_id=model_version_id,
            configuration_id=configuration_id,
        )


@job("high", timeout=settings.RQ_TIMEOUTS["worker_results_delete"])
def worker_results_delete(
        corpus_id: str,
        worker_run_id: Optional[str] = None,
        version_id: Optional[str] = None,
        element_id: Optional[str] = None,
        model_version_id: Optional[str] = None,
        configuration_id: Optional[str | Literal[False]] = None,
        include_children: bool = True) -> None:
    """
    Recursively delete all Worker Results produced by any WorkerVersion or a specific one on a
    whole corpus, under a specified parent element (parent element included), or on a single element.
    Results can be filtered depending on a specific model version and a specific or unset configuration.
    """
    assert (not worker_run_id or not version_id), "The worker_run_id and version_id parameters are mutually exclusive."

    _has_children = False
    if element_id:
        # Skip the child element deletion entirely if there are no ElementPaths with this element_id.
        # Setting this to False is equivalent to skipping all the way to the end of this task,
        # where we call the task again with include_children=False to delete on the element itself instead of its children.
        # This can be a significant speedup as this avoids multiple large DELETE queries.
        if include_children:
            _has_children = ElementPath.objects.filter(path__overlap=[element_id]).exists()

        # When running with a parent element, we do not delete any entities,
        # as checking whether they are used in other elements is too costly.
        entities = Entity.objects.none()
        # There will therefore be no TranscriptionEntities to delete because of entities
        entity_transcription_entities = TranscriptionEntity.objects.none()

        # include_children causes a deletion *only* on the element's descendants.
        # To also delete on the element itself, we will call this task synchronously with the same arguments and include_children=False.
        # This is used to avoid filtering by Q(id=element_id) | Q(paths__path__overlap=[element_id]) all at once,
        # which was found to be up to 30× slower than running two separate deletions.
        # When the element does not have any descendants, we skip this part entirely, behaving like include_children=False.
        if include_children and _has_children:
            elements = Element.objects.filter(paths__path__overlap=[element_id])
            classifications = Classification.objects.filter(element__paths__path__overlap=[element_id])
            transcriptions = Transcription.objects.filter(element__paths__path__overlap=[element_id])
            # TranscriptionEntities that will be removed due to the removal of transcriptions
            transcription_entities = TranscriptionEntity.objects.filter(transcription__element__paths__path__overlap=[element_id])
            metadata = MetaData.objects.filter(element__paths__path__overlap=[element_id])
            worker_activities = WorkerActivity.objects.filter(element__paths__path__overlap=[element_id])
        else:
            elements = Element.objects.filter(id=element_id)
            classifications = Classification.objects.filter(element_id=element_id)
            transcriptions = Transcription.objects.filter(element_id=element_id)
            # TranscriptionEntities that will be removed due to the removal of transcriptions
            transcription_entities = TranscriptionEntity.objects.filter(transcription__element_id=element_id)
            metadata = MetaData.objects.filter(element_id=element_id)
            worker_activities = WorkerActivity.objects.filter(element_id=element_id)

    else:
        elements = Element.objects.filter(corpus_id=corpus_id)
        classifications = Classification.objects.filter(element__corpus_id=corpus_id)
        transcriptions = Transcription.objects.filter(element__corpus_id=corpus_id)
        # TranscriptionEntities that will be removed due to the removal of transcriptions
        transcription_entities = TranscriptionEntity.objects.filter(transcription__element__corpus_id=corpus_id)
        # TranscriptionEntities that will be removed due to the removal of entities
        entity_transcription_entities = TranscriptionEntity.objects.filter(entity__corpus_id=corpus_id)
        entities = Entity.objects.filter(corpus_id=corpus_id)
        metadata = MetaData.objects.filter(element__corpus_id=corpus_id)
        worker_activities = WorkerActivity.objects.filter(element__corpus_id=corpus_id)

    # When a worker run ID is defined, filter by that worker run ID
    if worker_run_id:
        elements = elements.filter(worker_run_id=worker_run_id)
        classifications = classifications.filter(worker_run_id=worker_run_id)
        transcriptions = transcriptions.filter(worker_run_id=worker_run_id)
        transcription_entities = transcription_entities.filter(transcription__worker_run_id=worker_run_id)
        entity_transcription_entities = entity_transcription_entities.filter(worker_run_id=worker_run_id)
        entities = entities.filter(worker_run_id=worker_run_id)
        metadata = metadata.filter(worker_run_id=worker_run_id)
        # There is no worker_run_id on Worker Activities so the best thing we can do is delete the worker activities
        # attached to the elements produced with that worker run, and they are already being deleted by elements.trash()
        worker_activities = worker_activities.none()
    # When a version ID is defined, filter by the exact version ID
    elif version_id:
        elements = elements.filter(worker_version_id=version_id)
        classifications = classifications.filter(worker_version_id=version_id)
        transcriptions = transcriptions.filter(worker_version_id=version_id)
        transcription_entities = transcription_entities.filter(transcription__worker_version_id=version_id)
        entity_transcription_entities = entity_transcription_entities.filter(worker_version_id=version_id)
        entities = entities.filter(worker_version_id=version_id)
        metadata = metadata.filter(worker_version_id=version_id)
        worker_activities = worker_activities.filter(worker_version_id=version_id)
    # Otherwise, select everything that has any worker version ID. (When something has been created
    # by a worker run, it always has a worker version; however we have things that were created with
    # a worker version but without a worker run.)
    # We use worker_version_id != None and not worker_version_id__isnull=False,
    # because isnull would cause an unnecessary LEFT JOIN query.
    # No extra filtering is needed on worker activities, since worker versions cannot be null there.
    else:
        elements = elements.exclude(worker_version_id=None)
        classifications = classifications.exclude(worker_version_id=None)
        transcriptions = transcriptions.exclude(worker_version_id=None)
        transcription_entities = transcription_entities.exclude(transcription__worker_version_id=None)
        entity_transcription_entities = entity_transcription_entities.exclude(worker_version_id=None)
        entities = entities.exclude(worker_version_id=None)
        metadata = metadata.exclude(worker_version_id=None)

    if not worker_run_id and model_version_id:
        elements = elements.filter(worker_run__model_version_id=model_version_id)
        classifications = classifications.filter(worker_run__model_version_id=model_version_id)
        transcriptions = transcriptions.filter(worker_run__model_version_id=model_version_id)
        transcription_entities = transcription_entities.filter(transcription__worker_run__model_version_id=model_version_id)
        entity_transcription_entities = entity_transcription_entities.filter(worker_run__model_version_id=model_version_id)
        entities = entities.filter(worker_run__model_version_id=model_version_id)
        metadata = metadata.filter(worker_run__model_version_id=model_version_id)
        worker_activities = worker_activities.filter(model_version_id=model_version_id)

    if not worker_run_id and configuration_id is not None:
        if configuration_id is False:
            # Only delete results generated on a worker run with no configuration
            elements = elements.filter(worker_run__configuration_id=None)
            classifications = classifications.filter(worker_run__configuration_id=None)
            transcriptions = transcriptions.filter(worker_run__configuration_id=None)
            transcription_entities = transcription_entities.filter(transcription__worker_run__configuration_id=None)
            entity_transcription_entities = entity_transcription_entities.filter(worker_run__configuration_id=None)
            entities = entities.filter(worker_run__configuration_id=None)
            metadata = metadata.filter(worker_run__configuration_id=None)
            worker_activities = worker_activities.filter(configuration_id=None)
        else:
            elements = elements.filter(worker_run__configuration_id=configuration_id)
            classifications = classifications.filter(worker_run__configuration_id=configuration_id)
            transcriptions = transcriptions.filter(worker_run__configuration_id=configuration_id)
            transcription_entities = transcription_entities.filter(transcription__worker_run__configuration_id=configuration_id)
            entity_transcription_entities = entity_transcription_entities.filter(worker_run__configuration_id=configuration_id)
            entities = entities.filter(worker_run__configuration_id=configuration_id)
            metadata = metadata.filter(worker_run__configuration_id=configuration_id)
            worker_activities = worker_activities.filter(configuration_id=configuration_id)

    elements.trash()
    classifications.delete()

    # Delete TranscriptionEntities before transcriptions so that we can delete transcriptions using a single DELETE query
    transcription_entities.delete()
    entity_transcription_entities.delete()
    transcriptions._raw_delete(using="default")

    metadata.delete()

    # Remove entities from metadata and remove their associated transcription entities first
    # so that entities can be deleted in one query.
    MetaData.objects.filter(entity__in=entities).update(entity_id=None)
    TranscriptionEntity.objects.filter(entity__in=entities).delete()
    entities._raw_delete(using="default")

    worker_activities.delete()

    if element_id and include_children and _has_children:
        # Delete worker results on the parent element itself only after the deletion on its descendants has been done.
        # This deletion could cause the parent element to be deleted if it is considered to be a worker result itself,
        # which would remove the related ElementPaths and cause us to loose track of which child elements
        # we were supposed to delete worker results on.
        # This step is not needed if the element has no descendants, even when the function was called with include_children=True,
        # because the descendants deletion got skipped and the parent element has already been deleted.
        worker_results_delete(
            corpus_id=corpus_id,
            worker_run_id=worker_run_id,
            version_id=version_id,
            element_id=element_id,
            model_version_id=model_version_id,
            configuration_id=configuration_id,
            include_children=False,
        )


@job("high", timeout=settings.RQ_TIMEOUTS["move_element"])
def move_element(source: Element, destination: Element) -> None:
    parents = Element.objects.filter(id__in=source.paths.values("path__last")).only("id")
    for parent in parents:
        parent.remove_child(source)
    source.add_parent(destination)


@job("high", timeout=settings.RQ_TIMEOUTS["move_element"])
def move_selection(corpus_id: UUID, destination: Element) -> None:
    rq_job = get_current_job()
    assert rq_job is not None, "This task can only be run in a RQ job context."
    assert rq_job.user_id is not None, "This task requires a user ID to be defined on the RQ job."

    queryset = Element.objects.filter(selections__user_id=rq_job.user_id, corpus_id=corpus_id)
    total = queryset.count()
    for i, item in enumerate(queryset):
        rq_job.set_progress(i / total)
        move_element(source=item, destination=destination)


@job("high", timeout=settings.RQ_TIMEOUTS["move_element"])
def add_parent_selection(corpus_id: UUID, parent: Element) -> None:
    rq_job = get_current_job()
    assert rq_job is not None, "This task can only be run in a RQ job context."
    assert rq_job.user_id is not None, "This task requires a user ID to be defined on the RQ job."

    queryset = Element.objects.filter(selections__user_id=rq_job.user_id, corpus_id=corpus_id)
    total = queryset.count()
    for i, item in enumerate(queryset):
        rq_job.set_progress(i / total)
        item.add_parent(parent)


@job("default", timeout=settings.RQ_TIMEOUTS["reindex_corpus"])
def reindex_corpus(corpus_id: UUID, drop: bool = True) -> None:
    rq_job = get_current_job()
    assert rq_job is not None, "This task can only be run in a RQ job context."
    assert rq_job.user_id is not None, "This task requires a user ID to be defined on the RQ job."

    indexer = Indexer(corpus_id)
    if drop:
        indexer.drop_index()
    indexer.setup()
    indexer.index()

    # Report to the user that the index build finished
    user = User.objects.get(id=rq_job.user_id)
    corpus = Corpus.objects.get(id=corpus_id)
    send_mail(
        subject=f"Project {corpus.name} was successfully indexed",
        message=render_to_string(
            "reindex_corpus.html",
            context={
                "user": user,
                "corpus": corpus,
            },
        ),
        from_email=None,
        recipient_list=[user.email],
        fail_silently=True,
    )
