import json
import sqlite3
import tempfile
from pathlib import Path
from unittest.mock import patch

from django.contrib.gis.geos.linestring import LinearRing
from django.core.management import CommandError, call_command

from arkindex.documents.export import export_corpus
from arkindex.documents.management.commands.load_export import Command
from arkindex.documents.models import Corpus, Element, ElementPath, EntityType, Transcription
from arkindex.documents.tasks import corpus_delete
from arkindex.images.models import Image, ImageServer
from arkindex.process.models import ProcessMode, Worker, WorkerRun, WorkerType, WorkerVersion
from arkindex.project.tests import FixtureTestCase, force_constraints_immediate
from arkindex.training.models import Dataset, DatasetElement

BASE_DIR = Path(__file__).absolute().parent


class TestLoadExport(FixtureTestCase):

    def clean_dump_data(self, path, **field_overrides):
        """
        Clean the output of the dumpdata command to allow assertions.
        Extra keyword arguments can be used to allow overriding specific field values to handle changes in IDs:
        clean_dump_data('/tmp/lol.json', server={1: 2}) updates the field named `server` in any model,
        changing its value to 2 if it was 1 and leaving all other values unchanged.
        """
        # Some fields like id, created and updated are not comparable.
        # We keep only the comparable fields for each imported model.

        unexpected_fields_by_model = {
            "documents.elementtype": ["display_name", "indexable"],
            "documents.mlclass": [],
            "process.worker": [],
            "process.revision": ["message", "author"],
            "process.workerversion": ["created", "updated", "configuration", "state", "docker_image_iid"],
            # The WorkerRuns lose their parents, use different worker versions that just got recreated,
            # are assigned to the user's local process and not the original one
            "process.workerrun": ["parents", "version", "process", "summary", "created", "updated"],
            "process.workertype": [],
            "images.imageserver": ["s3_bucket", "s3_region", "created", "updated", "read_only"],
            "images.image": ["created", "updated", "hash", "status"],
            "documents.element": ["created", "updated", "type"],
            "documents.elementpath": [],
            "documents.entity": [],
            "documents.entitytype": [],
            "documents.transcription": [],
            "documents.transcriptionentity": [],
            "documents.metadata": ["index"],
            "documents.classification": ["ml_class"],
        }

        with open(path) as file:
            data = json.loads(file.read())

        results = []
        for object in data:
            if object["model"] not in unexpected_fields_by_model:
                continue

            unexpected_fields = unexpected_fields_by_model[object["model"]]
            current_fields = list(object["fields"].keys())

            # Delete the id
            del object["pk"]

            for field in current_fields:
                # Delete unexpected field
                if field in unexpected_fields:
                    del object["fields"][field]
                # Allow overriding fields
                elif object["fields"][field] in field_overrides.get(field, {}):
                    object["fields"][field] = field_overrides[field][object["fields"][field]]

            results.append(object)

        return results

    def test_invalid_file(self):
        with self.assertRaises(CommandError) as context:
            call_command("load_export", "wrong.txt", "--email", self.user.email)
        self.assertEqual(str(context.exception), "File wrong.txt is not an SQLite database")

    def test_file_not_exists(self):
        with self.assertRaises(CommandError) as context:
            call_command("load_export", "wrong.db", "--email", self.user.email)
        self.assertEqual(str(context.exception), "File wrong.db does not exist")

    def test_invalid_email(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        with self.assertRaises(CommandError) as context:
            call_command("load_export", temp_file, "--email", "wrong@email")
        self.assertEqual(str(context.exception), "User with the email wrong@email does not exist")

    def test_invalid_database(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        with self.assertRaises(CommandError) as context:
            call_command("load_export", temp_file, "--email", self.user.email)
        self.assertEqual(str(context.exception), (
            f"The SQLite database {temp_file} is missing some expected tables: "
            "['classification', 'dataset', 'dataset_element', 'element', "
            "'element_path', 'entity', 'entity_type', 'export_version', "
            "'image', 'image_server', 'metadata', 'transcription', "
            "'transcription_entity', 'worker_run', 'worker_version']"
        ))

    def test_invalid_version(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        db = sqlite3.connect(temp_file)
        cursor = db.cursor()
        cursor.executescript((BASE_DIR / "../../export/structure.sql").read_text())
        cursor.execute("UPDATE export_version SET version = 1")
        db.commit()
        db.close()

        with self.assertRaises(CommandError) as context:
            call_command("load_export", temp_file, "--email", self.user.email, "--corpus-name", "My corpus")
        self.assertEqual(str(context.exception), f"The SQLite database {temp_file} does not have a supported export version")

    @patch("arkindex.documents.export.os.unlink")
    @patch("arkindex.project.aws.s3.Object")
    def test_run(self, s3_object_mock, unlink_mock):
        Element.objects.filter(type__folder=False, image__isnull=True).update(
            polygon=LinearRing((0, 0), (0, 1000), (1000, 1000), (1000, 0), (0, 0)),
            image=Image.objects.all().first()
        )

        element = self.corpus.elements.get(name="Volume 1")
        transcription = Transcription.objects.first()

        reco_version = WorkerVersion.objects.get(worker__slug="reco")
        reco_run = reco_version.worker_runs.get(process__mode=ProcessMode.Workers)
        dla_version = WorkerVersion.objects.get(worker__slug="dla")
        dla_run = dla_version.worker_runs.get(process__mode=ProcessMode.Workers)
        init_version = WorkerVersion.objects.get(worker__slug="initialisation")

        dataset_set = Dataset.objects.first().sets.first()
        DatasetElement.objects.create(set=dataset_set, element=element)

        element.classifications.create(
            ml_class=self.corpus.ml_classes.create(name="Blah"),
            confidence=.55555555,
            worker_version=dla_version,
            worker_run=dla_run,
        )

        person_type = EntityType.objects.get(
            name="person",
            corpus=self.corpus
        )
        location_type = EntityType.objects.get(
            name="location",
            corpus=self.corpus
        )

        entity1 = self.corpus.entities.create(
            name="Arrokuda",
            type=location_type
        )
        entity2 = self.corpus.entities.create(
            name="Stonjourner",
            type=person_type,
            validated=True,
            moderator=self.superuser,
        )

        transcription.transcription_entities.create(
            entity=entity1,
            offset=1,
            length=1,
            worker_version=reco_version,
            worker_run=reco_run,
        )

        transcription.transcription_entities.create(
            entity=entity2,
            offset=0,
            length=1,
            worker_version=reco_version,
            worker_run=reco_run,
            confidence=0.42,
        )

        export = self.corpus.exports.create(user=self.user)

        export_corpus(export)

        # Retrieve the database path from the S3 upload argument
        args, _ = s3_object_mock().upload_file.call_args
        db_path = args[0]

        # Clean up WorkerRuns from local processes so we can check new worker runs are created
        WorkerRun.objects.filter(process__mode=ProcessMode.Local).delete()

        # Delete the elements initialisation worker run before calling dumpdata as it is not exported
        init_version = WorkerVersion.objects.get(worker__slug="initialisation")
        init_version.worker_runs.get().delete()

        # Call dumpdata command before the deletion
        # Ignore django_rq as it uses a fake database table to insert itself into Django's permissions system
        _, dump_path_before = tempfile.mkstemp(suffix=".json")
        call_command("dumpdata", output=dump_path_before, exclude=["django_rq"])

        # Delete the existing corpus
        corpus_delete(self.corpus.id)
        Image.objects.all().delete()
        ImageServer.objects.all().delete()
        WorkerVersion.objects.filter(id=reco_version.id).delete()

        with force_constraints_immediate():
            call_command("load_export", db_path, "--email", self.user.email, "--corpus-name", "My corpus")

        # Call dumpdata command after the import
        _, dump_path_after = tempfile.mkstemp(suffix=".json")
        call_command("dumpdata", output=dump_path_after, exclude=["django_rq"])

        corpus = Corpus.objects.get(name="My corpus")

        local_process = self.user.processes.get(mode=ProcessMode.Local)

        data_before = self.clean_dump_data(
            dump_path_before,
            # Remap the corpus, imageserver and worker version IDs
            corpus={str(self.corpus.id): str(corpus.id)},
            server={self.imgsrv.id: ImageServer.objects.get().id},
            worker_version={
                str(reco_version.id): str(WorkerVersion.objects.get(worker__slug="reco").id),
                str(dla_version.id): str(WorkerVersion.objects.get(worker__slug="dla").id),
            },
            worker_run={
                str(reco_run.id): str(local_process.worker_runs.get(version__worker__slug="reco").id),
                str(dla_run.id): str(local_process.worker_runs.get(version__worker__slug="dla").id),
            },
        )
        data_after = self.clean_dump_data(dump_path_after)

        self.assertCountEqual(data_before, data_after)

    def test_run_empty_database(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        db = sqlite3.connect(temp_file)
        cursor = db.cursor()
        cursor.executescript((BASE_DIR / "../../export/structure.sql").read_text())
        db.commit()
        db.close()

        with force_constraints_immediate():
            call_command("load_export", temp_file, "--email", self.user.email, "--corpus-name", "My corpus")

        corpus = Corpus.objects.get(name="My corpus")
        self.assertEqual(corpus.types.all().count(), 0)
        self.assertEqual(corpus.ml_classes.all().count(), 0)
        self.assertEqual(corpus.elements.all().count(), 0)
        self.assertEqual(corpus.entities.all().count(), 0)

    @patch("arkindex.documents.export.os.unlink")
    @patch("arkindex.project.aws.s3.Object")
    def test_run_objects_already_exist(self, s3_object_mock, unlink_mock):
        Element.objects.filter(type__folder=False, image__isnull=True).update(
            polygon=LinearRing((0, 0), (0, 1000), (1000, 1000), (1000, 0), (0, 0)),
            image=Image.objects.all().first()
        )

        element = self.corpus.elements.get(name="Volume 1")
        transcription = Transcription.objects.first()
        version = WorkerVersion.objects.get(worker__slug="reco")
        worker_run = version.worker_runs.first()

        element.classifications.create(
            ml_class=self.corpus.ml_classes.create(name="Blah"),
            confidence=.55555555,
        )

        dataset_set = Dataset.objects.first().sets.first()
        DatasetElement.objects.create(set=dataset_set, element=element)

        location_type = EntityType.objects.get(
            name="location",
            corpus=self.corpus
        )

        entity1 = self.corpus.entities.create(
            name="Arrokuda",
            type=location_type,
        )

        transcription.transcription_entities.create(
            entity=entity1,
            offset=1,
            length=1,
            worker_version=version,
            worker_run=worker_run,
        )

        export = self.corpus.exports.create(user=self.user)

        export_corpus(export)

        # Retrieve the database path from the S3 upload argument
        args, _ = s3_object_mock().upload_file.call_args
        db_path = args[0]

        # Remove the ElementPaths, because conflicts cannot be ignored with those.
        # There would normally be an error here, and the command would be able to keep on running,
        # but in this unit test context we are inside a transaction.
        # Any of the raw SQL queries, when they fail, will cause the whole transaction to be broken,
        # meaning the test can only fail.
        ElementPath.objects.filter(element__corpus=self.corpus).delete()

        with force_constraints_immediate():
            call_command("load_export", db_path, "--email", self.user.email, "--corpus-name", "My corpus")

        corpus = Corpus.objects.get(name="My corpus")
        self.assertEqual(corpus.types.all().count(), 6)
        self.assertEqual(corpus.ml_classes.all().count(), 1)
        self.assertEqual(corpus.elements.all().count(), 0)
        self.assertEqual(corpus.entities.all().count(), 0)

    def test_create_worker(self):
        command = Command()
        command.worker_type_map = {
            "recognizer": WorkerType.objects.get(slug="recognizer").id
        }
        row = {
            "slug": "some_worker",
            "repository_url": "http://my_repo.fake/workers/worker",
            "type": "recognizer",
            "name": "Some Worker"
        }
        command.create_worker(row)
        created_worker = Worker.objects.get(slug="some_worker")
        self.assertEqual(created_worker.name, "Some Worker")
        self.assertEqual(created_worker.type.slug, "recognizer")
