from io import StringIO
from textwrap import dedent

from django.core.management import call_command

from arkindex.documents.models import Entity, EntityType, MetaType
from arkindex.process.models import WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureTestCase


class TestMergeEntities(FixtureTestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.person_type = EntityType.objects.get(name="person", corpus=cls.corpus)
        cls.location_type = EntityType.objects.get(name="location", corpus=cls.corpus)

    def merge_entities(self):
        """
        Run the command, capturing its output
        """
        output = StringIO()
        call_command("merge_entities", corpus=self.corpus, no_color=True, stdout=output, stderr=output)
        return output.getvalue().strip()

    def test_merge_entities_nothing(self):
        self.assertEqual(self.corpus.entities.count(), 0)
        self.assertEqual(
            self.merge_entities(),
            dedent("""
            Looking for duplicate entities in Unit Tests…
            0 duplicate entities found.
            """).strip()
        )

    def test_merge_entities_no_duplicates(self):
        self.corpus.entities.create(name="A", type=self.person_type)
        self.corpus.entities.create(name="B", type=self.person_type)
        self.corpus.entities.create(name="B", type=self.location_type)
        self.assertEqual(
            self.merge_entities(),
            dedent("""
            Looking for duplicate entities in Unit Tests…
            0 duplicate entities found.
            """).strip()
        )

    def test_merge_entities(self):
        entity1 = self.corpus.entities.create(
            # Using custom UUIDs here because merge_entities picks which entity to keep
            # and which to remove by ordering on the ID; this keeps the tests deterministic.
            id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            name="A",
            type=self.person_type,
            metas={"a": "b"},
        )
        entity2 = self.corpus.entities.create(
            id="bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb",
            name="A",
            type=self.person_type,
            metas={"a": "c"},
        )
        unique_entity = self.corpus.entities.create(
            id="cccccccc-cccc-cccc-cccc-cccccccccccc",
            name="B",
            type=self.person_type,
        )

        element = self.corpus.elements.get(name="Volume 1, page 1r")
        element.metadatas.all().delete()
        element.metadatas.create(type=MetaType.Text, name="something", value="unique", entity=unique_entity)
        element.metadatas.create(type=MetaType.Text, name="something", value="dupe", entity=entity1)
        element.metadatas.create(type=MetaType.Text, name="something", value="dupe", entity=entity2)

        transcription = element.transcriptions.get()
        worker_version = WorkerVersion.objects.first()
        worker_version_2 = WorkerVersion.objects.first()
        worker_run = WorkerRun.objects.first()
        worker_run_2 = WorkerRun.objects.last()

        # Duplicate manual TranscriptionEntities
        transcription.transcription_entities.create(
            id=100,
            entity=entity1,
            offset=0,
            length=1,
        )
        transcription.transcription_entities.create(
            id=101,
            entity=entity2,
            offset=0,
            length=1,
        )
        # Duplicate TranscriptionEntities with worker versions and runs
        transcription.transcription_entities.create(
            id=102,
            entity=entity1,
            offset=0,
            length=1,
            worker_version=worker_version,
            worker_run=worker_run,
        )
        transcription.transcription_entities.create(
            id=103,
            entity=entity2,
            offset=0,
            length=1,
            worker_version=worker_version_2,
            worker_run=worker_run_2,
        )
        # Different offset: this should not be deduplicated.
        transcription.transcription_entities.create(
            id=106,
            entity=entity2,
            offset=1,
            length=1,
        )

        self.assertEqual(
            self.merge_entities(),
            dedent("""
            Looking for duplicate entities in Unit Tests…
            1 duplicate entities found.
            Updating entity IDs on metadata…
            Updated 1 metadata.
            Deleting duplicate TranscriptionEntities with worker runs…
            Deleted 0 TranscriptionEntities.
            Deleting duplicate manual TranscriptionEntities…
            Deleted 1 TranscriptionEntities.
            Updating entity IDs on TranscriptionEntities…
            Updated 2 TranscriptionEntities.
            Removing duplicate entities…
            Removed 1 duplicate entities from corpus Unit Tests.
            """).strip()
        )

        entity1.refresh_from_db()
        unique_entity.refresh_from_db()
        with self.assertRaises(Entity.DoesNotExist):
            # This entity got deleted
            entity2.refresh_from_db()

        self.assertListEqual(
            list(element.metadatas.order_by("value").values("name", "value", "entity_id")),
            [
                # Metadata are not deduplicated, only updated
                {
                    "name": "something",
                    "value": "dupe",
                    "entity_id": entity1.id,
                },
                {
                    "name": "something",
                    "value": "dupe",
                    "entity_id": entity1.id,
                },
                {
                    "name": "something",
                    "value": "unique",
                    "entity_id": unique_entity.id,
                },
            ]
        )

        self.assertListEqual(
            list(transcription.transcription_entities.order_by("id").values("entity_id", "offset", "length", "worker_version_id", "worker_run_id")),
            [
                {
                    "entity_id": entity1.id,
                    "offset": 0,
                    "length": 1,
                    "worker_version_id": None,
                    "worker_run_id": None,
                },
                {
                    "entity_id": entity1.id,
                    "offset": 0,
                    "length": 1,
                    "worker_version_id": worker_version.id,
                    "worker_run_id": worker_run.id,
                },
                {
                    "entity_id": entity1.id,
                    "offset": 0,
                    "length": 1,
                    "worker_version_id": worker_version.id,
                    "worker_run_id": worker_run_2.id,
                },
                {
                    "entity_id": entity1.id,
                    "offset": 1,
                    "length": 1,
                    "worker_version_id": None,
                    "worker_run_id": None,
                },
            ]
        )
