from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Classification, Corpus, MLClass
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestClasses(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = cls.corpus.ml_classes.create(name="text")
        cls.cover = cls.corpus.ml_classes.create(name="cover")

        cls.classified = cls.corpus.types.create(slug="classified", folder=True)
        cls.folder_type = cls.corpus.types.create(slug="folder", folder=True)

        cls.parent = cls.corpus.elements.create(type=cls.folder_type, name="parent")
        cls.common_children = cls.corpus.elements.create(type=cls.folder_type, name="common_children")

        cls.version1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.version2 = WorkerVersion.objects.get(worker__slug="dla")

        cls.worker_run_1 = cls.version1.worker_runs.first()
        cls.worker_run_2 = cls.version2.worker_runs.first()

        for elt_num in range(1, 13):
            elt = cls.corpus.elements.create(
                name=f"elt_{elt_num}",
                type=cls.classified,
            )
            elt.add_parent(cls.parent)
            cls.common_children.add_parent(elt)
            for ml_class, score in ((cls.text, .7), (cls.cover, .99)):
                for worker_run in (cls.worker_run_1, cls.worker_run_2):
                    elt.classifications.create(
                        worker_version=worker_run.version,
                        worker_run=worker_run,
                        ml_class=ml_class,
                        confidence=score,
                        high_confidence=bool(score == .99)
                    )

    def test_list(self):
        """
        Test listing results alpha-ordered
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.cover.pk),
                    "name": "cover",
                },
                {
                    "id": str(self.text.pk),
                    "name": "text",
                },
            ]
        })

        # Add other classes to ensure ordering is preserved
        self.white_page = MLClass.objects.create(name="white-page", corpus=self.corpus)
        self.image = MLClass.objects.create(name="image", corpus=self.corpus)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 4,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.cover.pk),
                    "name": "cover",
                },
                {
                    "id": str(self.image.pk),
                    "name": "image",
                },
                {
                    "id": str(self.text.pk),
                    "name": "text",
                },
                {
                    "id": str(self.white_page.pk),
                    "name": "white-page",
                },
            ]
        })

    def test_list_pagination_40(self):
        for i in range(40):
            MLClass.objects.create(name=f"class_{i}", corpus=self.corpus)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}), data={"page": 1})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["count"], 42)
        self.assertEqual(len(response.json()["results"]), 40)

    def test_unauthorized_methods(self):
        """
        Test unauthorized HTTP methods on ListCorpusMLClasses endpoint returns 405
        """
        self.client.force_login(self.user)
        for method in (self.client.put, self.client.patch, self.client.delete):
            response = method(reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}), {})
            self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_list_classes_search(self):
        buffer_class = MLClass.objects.create(name="buffer overflow", corpus=self.corpus)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}),
                data={"search": "over"}
            )
        data = response.json()
        self.assertEqual(data["count"], 2)
        self.assertListEqual(data["results"], [
            {
                "id": str(buffer_class.id),
                "name": "buffer overflow",
            }, {
                "id": str(self.cover.id),
                "name": "cover",
            }
        ])
        response = self.client.get(
            reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}),
            data={"search": "over buffer"}
        )
        self.assertListEqual(response.json()["results"], [{
            "id": str(buffer_class.id),
            "name": "buffer overflow",
        }])

    @override_settings(SEARCH_FILTER_MAX_TERMS=3)
    def test_search_filter_limit(self):
        self.corpus.ml_classes.create(name="buffer overflow")
        self.client.force_login(self.user)

        response = self.client.get(
            reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}),
            data={"search": "buf buf over buf flow"}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response = self.client.get(
            reverse("api:corpus-classes", kwargs={"pk": self.corpus.pk}),
            data={"search": "buf buff over buf flow"}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"search": ["There cannot be more than 3 unique search terms."]})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_corpus_classes_corpus_rights(self, has_access_mock):
        self.client.force_login(self.user)
        private_corpus = Corpus.objects.create(name="private")
        response = self.client.post(reverse("api:corpus-classes", kwargs={"pk": private_corpus.pk}), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_classes_create_requires_login(self):
        response = self.client.post(reverse("api:corpus-classes", kwargs={"pk": self.corpus.id}), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_classes_create(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:corpus-classes", kwargs={"pk": self.corpus.id}), {"name": "a"})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_class = self.corpus.ml_classes.get(name="a")
        self.assertDictEqual(response.json(), {
            "id": str(new_class.id),
            "name": "a"
        })

    def test_corpus_classes_create_unique(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:corpus-classes", kwargs={"pk": self.corpus.id}), {"name": "text"})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "non_field_errors": ["The fields name, corpus must make a unique set."]
        })

    def test_retrieve(self):
        self.client.force_login(self.superuser)
        response = self.client.get(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.text.id),
            "name": "text"
        })

    def test_retrieve_requires_login(self):
        response = self.client.get(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update(self):
        self.client.force_login(self.superuser)
        response = self.client.put(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.text.id),
            "name": "new name"
        })
        self.text.refresh_from_db()
        self.assertEqual(self.text.name, "new name")

    def test_update_requires_login(self):
        response = self.client.put(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.put(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_update_requires_admin(self, has_access_mock):
        self.user.rights.update(level=Role.Contributor.value)
        self.client.force_login(self.user)
        response = self.client.put(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this corpus."})

    def test_partial_update(self):
        self.client.force_login(self.superuser)
        response = self.client.patch(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.text.id),
            "name": "new name"
        })
        self.text.refresh_from_db()
        self.assertEqual(self.text.name, "new name")

    def test_partial_update_requires_login(self):
        response = self.client.patch(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.patch(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_partial_update_requires_admin(self, has_access_mock):
        self.user.rights.update(level=Role.Contributor.value)
        self.client.force_login(self.user)
        response = self.client.patch(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}), {"name": "new name"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this corpus."})

    def test_destroy(self):
        self.client.force_login(self.superuser)
        self.text.classifications.all().delete()
        response = self.client.delete(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(MLClass.DoesNotExist):
            self.text.refresh_from_db()

    def test_destroy_non_empty(self):
        self.client.force_login(self.superuser)
        self.assertTrue(self.text.classifications.exists())
        response = self.client.delete(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["This ML class is used by some classifications on this corpus."])

    def test_destroy_requires_login(self):
        response = self.client.delete(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_destroy_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.delete(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_destroy_requires_admin(self, has_access_mock):
        self.user.rights.update(level=Role.Contributor.value)
        self.client.force_login(self.user)
        response = self.client.delete(reverse("api:ml-class-retrieve", kwargs={"corpus": self.corpus.id, "mlclass": self.text.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this corpus."})

    def test_list_elements_db_queries(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                data={"type": self.classified.slug}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["count"], 12)

    def test_list_elements_classes(self):
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                data={"type": self.classified.slug, "with_classes": True}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], 12)
        for elt in data["results"]:
            self.assertCountEqual(
                list(map(lambda c: (c["worker_run"], c["ml_class"]["name"], c["confidence"]), elt["classifications"])),
                [
                    ({"id": str(self.worker_run_1.id), "summary": self.worker_run_1.summary}, "cover", .99),
                    ({"id": str(self.worker_run_2.id), "summary": self.worker_run_2.summary}, "cover", .99),
                    ({"id": str(self.worker_run_1.id), "summary": self.worker_run_1.summary}, "text", .7),
                    ({"id": str(self.worker_run_2.id), "summary": self.worker_run_2.summary}, "text", .7),
                ]
            )

    def test_list_elements_classes_false(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                data={
                    "type": self.classified.slug,
                    "with_classes": False,
                }
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        for elt in response.json()["results"]:
            self.assertIsNone(elt["classifications"])

    def test_element_parents_classes(self):
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:elements-parents", kwargs={"pk": str(self.common_children.id)}),
                data={"type": self.classified.slug, "with_classes": 1}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], 12)
        for elt in data["results"]:
            self.assertCountEqual(
                list(map(lambda c: (c["worker_run"], c["ml_class"]["name"], c["confidence"]), elt["classifications"])),
                [
                    ({"id": str(self.worker_run_1.id), "summary": self.worker_run_1.summary}, "cover", .99),
                    ({"id": str(self.worker_run_2.id), "summary": self.worker_run_2.summary}, "cover", .99),
                    ({"id": str(self.worker_run_1.id), "summary": self.worker_run_1.summary}, "text", .7),
                    ({"id": str(self.worker_run_2.id), "summary": self.worker_run_2.summary}, "text", .7),
                ]
            )

    def test_element_children_classes(self):
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.parent.id)}),
                data={"type": self.classified.slug, "with_classes": "yes"}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], 12)
        for elt in data["results"]:
            self.assertCountEqual(
                list(map(lambda c: (c["worker_run"], c["ml_class"]["name"], c["confidence"]), elt["classifications"])),
                [
                    ({"id": str(self.worker_run_1.id), "summary": self.worker_run_1.summary}, "cover", .99),
                    ({"id": str(self.worker_run_2.id), "summary": self.worker_run_2.summary}, "cover", .99),
                    ({"id": str(self.worker_run_1.id), "summary": self.worker_run_1.summary}, "text", .7),
                    ({"id": str(self.worker_run_2.id), "summary": self.worker_run_2.summary}, "text", .7),
                ]
            )

    def test_element_lists_invalid_class_filters(self):
        corpus2 = Corpus.objects.create(name="Corpus 2")
        other_class = corpus2.ml_classes.create(name="something")

        endpoints = [
            reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
            reverse("api:elements-children", kwargs={"pk": str(self.parent.id)}),
            reverse("api:elements-parents", kwargs={"pk": str(self.common_children.id)}),
        ]
        cases = [
            ({"class_id": "lol"}, {"class_id": ["“lol” is not a valid UUID."]}),
            ({"class_id": str(other_class.id)}, {"class_id": [f'ML class "{other_class.id}" not found']}),
            ({"classification_confidence": "very"}, {"classification_confidence": ["could not convert string to float: 'very'"]}),
            ({"classification_confidence": "nan"}, {"classification_confidence": ["Confidence must be between 0 and 1"]}),
            ({"classification_confidence": "inf"}, {"classification_confidence": ["Confidence must be between 0 and 1"]}),
            ({"classification_confidence": "42"}, {"classification_confidence": ["Confidence must be between 0 and 1"]}),
            ({"classification_confidence": "-.5"}, {"classification_confidence": ["Confidence must be between 0 and 1"]}),
            (
                {"classification_confidence_operator": "hah"},
                {"classification_confidence_operator": [
                    "This operator is not supported.",
                    "This option is not supported without classification_confidence.",
                ]}
            ),
            (
                {"classification_confidence": "0.3", "classification_confidence_operator": "lol"},
                {"classification_confidence_operator": ["This operator is not supported."]}
            )
        ]
        for endpoint in endpoints:
            for data, expected_errors in cases:
                with self.subTest(endpoint=endpoint, **data):
                    response = self.client.get(endpoint, data)
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(response.json(), expected_errors)

    def test_element_lists_class_filters(self):
        # Create more diverse test data for these filters
        Classification.objects.all().delete()
        for i in range(1, 7):
            self.corpus.elements.get(name=f"elt_{i}").classifications.create(
                worker_version=self.version1,
                worker_run=self.worker_run_1,
                ml_class=self.text,
                confidence=i / 6,
                # Give every even element a high confidence classification
                high_confidence=i % 2 == 0
            )
        # Second half of elements gets self.cover
        for i in range(7, 13):
            self.corpus.elements.get(name=f"elt_{i}").classifications.create(
                worker_version=self.version2,
                worker_run=self.worker_run_2,
                ml_class=self.cover,
                confidence=(i - 6) / 6,
                high_confidence=i % 2 == 0
            )

        endpoints = [
            reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
            reverse("api:elements-children", kwargs={"pk": str(self.parent.id)}),
            reverse("api:elements-parents", kwargs={"pk": str(self.common_children.id)}),
        ]
        cases = [
            (
                {"class_id": str(self.text.id)},
                ["elt_1", "elt_2", "elt_3", "elt_4", "elt_5", "elt_6"],
            ),
            (
                {"classification_confidence": "0.5"},
                ["elt_3", "elt_9"],
            ),
            (
                {"classification_confidence": "0.51"},
                [],
            ),
            (
                {"classification_confidence": "0.5", "classification_confidence_operator": "gt"},
                ["elt_4", "elt_5", "elt_6", "elt_10", "elt_11", "elt_12"]
            ),
            (
                {"classification_confidence": "0.7", "classification_confidence_operator": "lte"},
                ["elt_1", "elt_2", "elt_3", "elt_4", "elt_7", "elt_8", "elt_9", "elt_10"],
            ),
            (
                {"class_id": str(self.cover.id), "classification_confidence": "0.7"},
                [],
            ),
            (
                {"classification_high_confidence": True},
                ["elt_2", "elt_4", "elt_6", "elt_8", "elt_10", "elt_12"],
            ),
            (
                {"classification_high_confidence": False},
                ["elt_1", "elt_3", "elt_5", "elt_7", "elt_9", "elt_11"],
            ),
            (
                {"class_id": str(self.text.id), "classification_high_confidence": True},
                ["elt_2", "elt_4", "elt_6"],
            ),
            (
                {"class_id": str(self.text.id), "classification_high_confidence": False},
                ["elt_1", "elt_3", "elt_5"],
            ),
        ]
        for endpoint in endpoints:
            for data, element_names in cases:
                with self.subTest(endpoint=endpoint, **data):
                    response = self.client.get(
                        endpoint,
                        data={**data, "type": self.classified.slug},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())
                    self.assertCountEqual(
                        [element["name"] for element in response.json()["results"]],
                        element_names,
                    )
