from datetime import datetime, timezone
from itertools import cycle
from unittest.mock import patch

from django.db import connections
from django.db.utils import IntegrityError
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.process.models import Process, WorkerActivity, WorkerActivityState, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_tree
from arkindex.training.models import Dataset
from arkindex.users.models import Role


class TestDestroyElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug="volume")
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.surface = cls.corpus.elements.get(name="Surface A")
        cls.private_corpus = Corpus.objects.create(name="private", public=False)

    def test_element_destroy_verified_user(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_element_destroy_acl(self, has_access_mock):
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        castle_story = self.private_corpus.elements.create(
            type=self.volume_type,
            name="Castle story"
        )
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:element-retrieve", kwargs={"pk": str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have admin access to this element."}
        )

    @patch("arkindex.project.triggers.documents_tasks.element_delete.delay")
    def test_element_destroy(self, delay_mock):
        self.client.force_login(self.user)
        castle_story = self.corpus.elements.create(
            type=self.volume_type,
            name="Castle story"
        )
        self.assertTrue(self.corpus.elements.filter(id=castle_story.id).exists())
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse(
                    "api:element-retrieve",
                    kwargs={"pk": str(castle_story.id)}
                ),
                query_params={"delete_children": "False"},
            )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertDictEqual(kwargs, {
            "element": castle_story,
            "delete_children": False,
            "user_id": self.user.id,
            "description": "Deletion of Volume: Castle story",
        })

    @patch("arkindex.project.triggers.documents_tasks.element_delete.delay")
    def test_element_destroy_delete_children(self, delay_mock):
        self.client.force_login(self.user)
        castle_story = self.corpus.elements.create(
            type=self.volume_type,
            name="Castle story"
        )
        self.assertTrue(self.corpus.elements.filter(id=castle_story.id).exists())

        cases = [
            (True, "Recursive deletion of Volume: Castle story"),
            (False, "Deletion of Volume: Castle story"),
        ]

        for delete_children, expected_description in cases:
            with self.subTest(delete_children=delete_children):
                delay_mock.reset_mock()

                with self.assertNumQueries(3):
                    response = self.client.delete(
                        reverse("api:element-retrieve", kwargs={"pk": str(castle_story.id)}),
                        query_params={"delete_children": str(delete_children)},
                    )
                self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

                self.assertEqual(delay_mock.call_count, 1)
                args, kwargs = delay_mock.call_args
                self.assertEqual(len(args), 0)
                self.assertDictEqual(kwargs, {
                    "element": castle_story,
                    "delete_children": delete_children,
                    "user_id": self.user.id,
                    "description": expected_description,
                })

    @patch("arkindex.project.triggers.documents_tasks.element_delete.delay")
    def test_element_destroy_creator(self, delay_mock):
        """
        An element's creator can delete the element if it has write access
        """
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        castle_story = self.private_corpus.elements.create(
            type=self.volume_type,
            name="Castle story",
            creator=self.user,
        )
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse(
                    "api:element-retrieve",
                    kwargs={"pk": str(castle_story.id)}
                ),
                query_params={"delete_children": "False"},
            )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertDictEqual(kwargs, {
            "element": castle_story,
            "delete_children": False,
            "user_id": self.user.id,
            "description": "Deletion of Volume: Castle story",
        })

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_element_destroy_creator_acl(self, has_access_mock):
        """
        An element's creator cannot delete without write access
        """
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        castle_story = self.private_corpus.elements.create(
            type=self.volume_type,
            name="Castle story",
            creator=self.user,
        )
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:element-retrieve", kwargs={"pk": str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have write access to this element."}
        )

    def test_element_destroy_in_dataset_forbidden(self):
        """
        An element cannot be deleted via the API if linked to a dataset
        """
        dataset = Dataset.objects.get(name="First Dataset")
        dataset.sets.get(name="test").set_elements.create(element=self.vol)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You cannot delete an element that is part of a dataset."})

    @patch("arkindex.project.triggers.documents_tasks.element_delete.delay")
    def test_non_empty_element(self, delay_mock):
        """
        We can now delete a non-empty element
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse(
                    "api:element-retrieve",
                    kwargs={"pk": str(self.vol.id)}
                ),
                query_params={"delete_children": "False"},
            )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertDictEqual(kwargs, {
            "element": self.vol,
            "delete_children": False,
            "user_id": self.user.id,
            "description": "Deletion of Volume: Volume 1",
        })

    def test_element_trash_dataset_failure(self):
        """
        Elements that are part of a dataset cannot be deleted
        """
        dataset = Dataset.objects.get(name="First Dataset")
        dataset.sets.get(name="test").set_elements.create(
            element=Element.objects.get_descending(self.vol.id).first()
        )

        Element.objects.filter(id=self.vol.id).trash()
        # Prevent delaying constraints check at end of the test transaction
        # by directly setting "SET CONSTRAINTS ALL IMMEDIATE"
        # https://code.djangoproject.com/ticket/11665
        with self.assertRaises(IntegrityError):
            connections["default"].check_constraints()

    def test_element_trash_children(self):
        # Check this volume has children
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list("id", flat=True))
        self.assertEqual(len(ids), 24)
        self.assertEqual(len(Element.objects.exclude(id__in=ids)), 5)

        with self.assertExactQueries("element_trash_children.sql", params={"id": self.vol.id.hex}):
            Element.objects.filter(id=self.vol.id).trash()

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        self.assertFalse(Element.objects.filter(id__in=ids).exists())
        # Excluding the children, only the volume should have been deleted
        self.assertEqual(len(Element.objects.exclude(id__in=ids)), 4)

    def test_element_trash_without_children(self):
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list("id", flat=True))
        self.assertEqual(len(ids), 24)

        with self.assertExactQueries("element_trash_no_children.sql", params={"id": self.vol.id.hex}):
            Element.objects.filter(id=self.vol.id).trash(delete_children=False)

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        # No children were deleted
        self.assertEqual(Element.objects.filter(id__in=ids).count(), 24)

    def test_element_trash_no_order_by(self):
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list("id", flat=True))
        self.assertEqual(len(ids), 24)

        with self.assertExactQueries("element_trash_children.sql", params={"id": self.vol.id.hex}):
            Element.objects.filter(id=self.vol.id).order_by("name").trash()

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        self.assertFalse(Element.objects.filter(id__in=ids).exists())

    def test_element_trash_related_filter_bug(self):
        """
        This test should fail when a bug in ElementQuerySet.trash is fixed:
        Filtering by metadata, classification, selection, ProcessElement, or Process
        can cause elements to not actually be deleted as the related tables are deleted first.
        For example, deleting all elements with classifications will cause all classifications
        to be deleted, but not elements, because the QuerySet is re-evaluated
        and no elements will have classifications.
        """
        ml_class = self.corpus.ml_classes.create(name="Krabby")
        page1r, page1v, page2r = Element.objects.get_descending(self.vol.id).filter(type__slug="page")
        page1r.classifications.create(ml_class=ml_class, confidence=0.42)
        page1v.classifications.create(ml_class=ml_class, confidence=0.42)
        page2r.classifications.create(ml_class=ml_class, confidence=0.42)

        queryset = Element.objects.filter(classifications__isnull=False)

        # Ensure the pages are all in this queryset, and therefore should be deleted
        self.assertEqual(queryset.filter(id__in=[page1r.id, page1v.id, page2r.id]).count(), 3)

        with self.assertExactQueries("element_trash_ml_class.sql"):
            Element.objects.filter(classifications__isnull=False).trash()

        # Pages did not get deleted
        page1r.refresh_from_db()
        page1v.refresh_from_db()
        page2r.refresh_from_db()

        # Classifications have been deleted
        self.assertFalse(ml_class.classifications.exists())

    def test_element_trash_removes_activities(self):
        """
        Worker Activities must be deleted when trashing elements
        """
        process = Process.objects.first()
        elements = Element.objects.get_descending(self.vol.id).filter(type__slug="page")
        self.assertEqual(elements.count(), 3)
        states = cycle(WorkerActivityState)

        WorkerActivity.objects.bulk_create(
            WorkerActivity(
                process=process,
                element_id=elt.id,
                worker_version_id=WorkerVersion.objects.get(worker__slug="reco").id,
                state=next(states),
                started=datetime.now(timezone.utc),
            ) for elt in elements
        )
        elements.trash()
        self.assertEqual(WorkerActivity.objects.filter(element_id__in=elements.values("id")).count(), 0)

    def test_element_trash_deep(self):
        """
        Tests the repetition of path deletions with deep hierarchies
        Builds a tree of elements A→B→C→D→E and deletes A with its children
        """
        elements = build_tree(
            {
                "B": "A",
                "C": "B",
                "D": "C",
                "E": "D",
            },
            type=self.corpus.types.create(slug="houndour", display_name="Houndour", folder=True),
            corpus=self.corpus,
        )

        with self.assertExactQueries("element_trash_deep.sql", params={"id": elements["A"].id.hex}):
            Element.objects.filter(id=elements["A"].id).trash()

        self.assertFalse(Element.objects.filter(id__in=[e.id for e in elements.values()]).exists())

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_corpus_elements_requires_login(self, delay_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_destroy_corpus_elements_requires_admin(self, has_access_mock, delay_mock):
        self.corpus.memberships.update(level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this corpus."})

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_corpus_elements_empty(self, delay_mock):
        self.client.force_login(self.user)
        self.assertFalse(self.corpus.elements.filter(name="blablablabla").exists())

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                QUERY_STRING="name=blablablabla",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_corpus_elements(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(list(kwargs.pop("queryset")), list(self.corpus.elements.all()))
        self.assertDictEqual(kwargs, {
            "delete_children": True,
            "user_id": self.user.id,
            "description": "Element deletion",
        })

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_corpus_elements_delete_children(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                QUERY_STRING="delete_children=false"
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(list(kwargs.pop("queryset")), list(self.corpus.elements.all()))
        self.assertDictEqual(kwargs, {
            "delete_children": False,
            "user_id": self.user.id,
            "description": "Element deletion",
        })

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_children_requires_login(self, delay_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:elements-children", kwargs={"pk": self.vol.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_destroy_element_children_requires_admin(self, has_access_mock, delay_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:elements-children", kwargs={"pk": self.vol.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this corpus."})

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_children_empty(self, delay_mock):
        self.client.force_login(self.user)
        element = self.corpus.elements.create(type=self.volume_type, name="Lonely element")

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:elements-children", kwargs={"pk": element.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_children(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:elements-children", kwargs={"pk": self.vol.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop("queryset")),
            # Direct children of the volume
            list(Element.objects.get_descending(self.vol.id).filter(paths__path__last=self.vol.id)),
        )
        self.assertDictEqual(kwargs, {
            "delete_children": True,
            "user_id": self.user.id,
            "description": "Element deletion",
        })

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_children_delete_children(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                QUERY_STRING="delete_children=false"
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop("queryset")),
            # Direct children of the volume
            list(Element.objects.get_descending(self.vol.id).filter(paths__path__last=self.vol.id)),
        )
        self.assertDictEqual(kwargs, {
            "delete_children": False,
            "user_id": self.user.id,
            "description": "Element deletion",
        })

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_parents_requires_login(self, delay_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:elements-parents", kwargs={"pk": self.surface.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_destroy_element_parents_requires_writable(self, has_access_mock, delay_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:elements-parents", kwargs={"pk": self.surface.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this corpus."})

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_parents_empty(self, delay_mock):
        self.client.force_login(self.user)
        element = self.corpus.elements.create(type=self.volume_type, name="Lonely element")

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:elements-parents", kwargs={"pk": element.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_parents(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:elements-parents", kwargs={"pk": self.surface.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop("queryset")),
            # Direct parents of the surface
            list(Element.objects.get_ascending(self.surface.id, recursive=False)),
        )
        self.assertDictEqual(kwargs, {
            "delete_children": True,
            "user_id": self.user.id,
            "description": "Element deletion",
        })

    @patch("arkindex.project.triggers.documents_tasks.element_trash.delay")
    def test_destroy_element_parents_delete_parents(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:elements-parents", kwargs={"pk": self.surface.id}),
                QUERY_STRING="delete_children=false"
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop("queryset")),
            # Direct parents of the surface
            list(Element.objects.get_ascending(self.surface.id, recursive=False)),
        )
        self.assertDictEqual(kwargs, {
            "delete_children": False,
            "user_id": self.user.id,
            "description": "Element deletion",
        })

    def test_element_dot_delete(self):
        """
        test Element.delete method
        """
        self.client.force_login(self.user)
        with self.assertExactQueries("element_dot_delete.sql", params={"id": self.vol.id.hex}):
            self.vol.delete()
        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()
