from unittest.mock import call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import Worker, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState


class TestDestroyWorkerResults(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.version = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_run = WorkerRun.objects.get(version=cls.version)
        cls.page = cls.corpus.elements.get(name="Volume 1, page 2r")
        cls.private_corpus = Corpus.objects.create(name="private", public=False)
        cls.model = Model.objects.create(name="Generic model", public=False)
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            tag="Test",
            hash="A" * 32,
            archive_hash="42",
            size=1337,
        )
        cls.configuration = Worker.objects.get(slug="dla").configurations.create(
            name="DLA configuration", configuration={"value": "test"}
        )

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_wrong_acl(self, has_access_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:worker-delete-results", kwargs={"corpus": str(self.private_corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have admin access to this corpus."}
        )

    def test_wrong_corpus_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:worker-delete-results", kwargs={"corpus": "12341234-1234-1234-1234-123412341234"}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(
            response.json(),
            {"detail": "No Corpus matches the given query."}
        )

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_no_filter(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            worker_run_id=None,
            version_id=None,
            element_id=None,
            model_version_id=None,
            configuration_id=None,
            user_id=self.user.id,
            description="Deletion of worker results",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_version(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"worker_version_id": str(self.version.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            worker_run_id=None,
            element_id=None,
            version_id=self.version.id,
            model_version_id=None,
            configuration_id=None,
            user_id=self.user.id,
            description=f"Deletion of worker results produced by {self.version}",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_worker_run_ignore_filters(self, delay_mock):
        """
        When worker_run_id is passed, worker_version_id, model_version_id and configuration_id
        are ignored
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={
                    "worker_version_id": str(self.version.id),
                    "worker_run_id": str(self.worker_run.id),
                    "model_version_id": str(self.model_version.id),
                    "configuration_id": "false",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=None,
            element_id=None,
            worker_run_id=self.worker_run.id,
            model_version_id=None,
            configuration_id=None,
            user_id=self.user.id,
            description=f"Deletion of worker results produced by {self.worker_run.summary}",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_worker_run(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"worker_run_id": str(self.worker_run.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=None,
            element_id=None,
            worker_run_id=self.worker_run.id,
            model_version_id=None,
            configuration_id=None,
            user_id=self.user.id,
            description=f"Deletion of worker results produced by {self.worker_run.summary}",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_element(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"element_id": str(self.page.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            element_id=self.page.id,
            worker_run_id=None,
            version_id=None,
            model_version_id=None,
            configuration_id=None,
            user_id=self.user.id,
            description="Deletion of worker results",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_element_worker_run(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={
                    "element_id": str(self.page.id),
                    "worker_run_id": str(self.worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            element_id=self.page.id,
            worker_run_id=self.worker_run.id,
            version_id=None,
            model_version_id=None,
            configuration_id=None,
            user_id=self.user.id,
            description=f"Deletion of worker results produced by {self.worker_run.summary}",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_unset_configuration(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"configuration_id": "false"},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            element_id=None,
            worker_run_id=None,
            version_id=None,
            model_version_id=None,
            configuration_id=False,
            user_id=self.user.id,
            description="Deletion of worker results with no configuration",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_model_version(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"model_version_id": str(self.model_version.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            element_id=None,
            worker_run_id=None,
            version_id=None,
            model_version_id=self.model_version.id,
            configuration_id=None,
            user_id=self.user.id,
            description=f"Deletion of worker results based on model version Generic model @ Test ({str(self.model_version.id)[:6]}…)",
        ))

    @patch("arkindex.project.triggers.documents_tasks.worker_results_delete.delay")
    def test_filter_element_worker_version_model_version_configuration(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={
                    "worker_version_id": str(self.version.id),
                    "element_id": str(self.page.id),
                    "model_version_id": str(self.model_version.id),
                    "configuration_id": str(self.configuration.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            element_id=self.page.id,
            worker_run_id=None,
            version_id=self.version.id,
            model_version_id=self.model_version.id,
            configuration_id=self.configuration.id,
            user_id=self.user.id,
            description=(
                "Deletion of worker results produced by Recognizer"
                f" for version 1 based on model version Generic model @ Test ({str(self.model_version.id)[:6]}…)"
                " configured with DLA configuration"
            )
        ))

    def test_invalid_version_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"worker_version_id": "lol"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"worker_version_id": ["Invalid UUID."]},
        )

    def test_wrong_version_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"worker_version_id": "12341234-1234-1234-1234-123412341234"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"worker_version_id": ["This worker version does not exist."]}
        )

    def test_invalid_worker_run_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"worker_run_id": "lol"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"worker_run_id": ["Invalid UUID."]},
        )

    def test_wrong_worker_run_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"worker_run_id": "12341234-1234-1234-1234-123412341234"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"worker_run_id": ["This worker run does not exist."]}
        )

    def test_invalid_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"element_id": "lol"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"element_id": ["Invalid UUID."]}
        )

    def test_wrong_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"element_id": "12341234-1234-1234-1234-123412341234"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"element_id": ["This element does not exist in the specified corpus."]}
        )

    def test_invalid_model_version_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"model_version_id": "lol"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"model_version_id": ["Invalid UUID."]}
        )

    def test_wrong_model_version_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"model_version_id": "12341234-1234-1234-1234-123412341234"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"model_version_id": ["This model version does not exist."]}
        )

    def test_invalid_configuration_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"configuration_id": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"configuration_id": ['Invalid UUID. You can set "false" to exclude results with a configuration.']}
        )

    def test_wrong_configuration_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"configuration_id": "12341234-1234-1234-1234-123412341234"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"configuration_id": ["This worker configuration does not exist."]}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": False})
    def test_selection_feature_flag(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"use_selection": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"use_selection": ["Selection is not available on this instance."]}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_selection_no_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={
                    "use_selection": "true",
                    "element_id": str(self.page.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"use_selection": [
                "use_selection and element_id cannot be used simultaneously.",
                "No elements of the specified corpus have been selected."
            ]}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_selection_empty(self):
        self.client.force_login(self.user)
        self.assertFalse(self.user.selected_elements.exists())
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"use_selection": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"use_selection": ["No elements of the specified corpus have been selected."]}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_selection_empty_for_corpus(self):
        self.superuser.selected_elements.add(self.page)
        self.assertFalse(self.user.selected_elements.filter(corpus=self.private_corpus).exists())
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.private_corpus.id)}),
                query_params={"use_selection": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(
                response.json(),
                {"use_selection": ["No elements of the specified corpus have been selected."]}
            )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    @patch("arkindex.project.triggers.documents_tasks.selection_worker_results_delete.delay")
    def test_selection(self, delay_mock):
        self.user.selected_elements.add(self.page)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={"use_selection": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            worker_run_id=None,
            version_id=None,
            user_id=self.user.id,
            model_version_id=None,
            configuration_id=None,
            description=f"Deletion of worker results on selected elements in {self.corpus.name}"
        ))

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    @patch("arkindex.project.triggers.documents_tasks.selection_worker_results_delete.delay")
    def test_selection_version_filter(self, delay_mock):
        self.user.selected_elements.add(self.page)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={
                    "use_selection": "true",
                    "worker_version_id": str(self.version.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            worker_run_id=None,
            version_id=self.version.id,
            user_id=self.user.id,
            model_version_id=None,
            configuration_id=None,
            description=f"Deletion of worker results on selected elements in {self.corpus.name} "
                        f"produced by {self.version}"
        ))

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    @patch("arkindex.project.triggers.documents_tasks.selection_worker_results_delete.delay")
    def test_selection_worker_run_filter(self, delay_mock):
        self.user.selected_elements.add(self.page)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:worker-delete-results", kwargs={"corpus": str(self.corpus.id)}),
                query_params={
                    "use_selection": "true",
                    "worker_run_id": str(self.worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            worker_run_id=self.worker_run.id,
            version_id=None,
            user_id=self.user.id,
            model_version_id=None,
            configuration_id=None,
            description=f"Deletion of worker results on selected elements in {self.corpus.name} "
                        f"produced by {self.worker_run.summary}"
        ))
