from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, TextOrientation
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestTranscriptions(FixtureAPITestCase):
    """
    Tests for text element view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.volume = cls.corpus.elements.get(name="Volume 1")
        cls.line = cls.corpus.elements.get(name="Text line")
        cls.private_corpus = Corpus.objects.create(name="Private")
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user("a@bc.de", "a")
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()
        cls.worker_version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.worker_run = WorkerRun.objects.filter(version=cls.worker_version_1, process__mode=ProcessMode.Workers).get()

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_list_transcriptions_read_right(self, filter_rights_mock):
        # A read right on the element corpus is required to access transcriptions
        self.client.force_login(self.private_read_user)

        response = self.client.get(reverse("api:element-transcriptions", kwargs={"pk": str(self.private_page.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.private_read_user, Corpus, Role.Guest.value))

    def test_list_element_transcriptions(self):
        tr1 = self.page.transcriptions.get()
        tr2 = self.page.transcriptions.create(
            text="something",
            worker_version=self.worker_version_2,
            worker_run=self.worker_run,
            confidence=0.369,
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual(response.json()["results"], [
            {
                "id": str(tr1.id),
                "text": "Lorem ipsum dolor sit amet",
                "confidence": 1.0,
                "orientation": TextOrientation.HorizontalLeftToRight.value,
                "element": None,
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": "Worker Recognizer @ version 1",
                },
            },
            {
                "id": str(tr2.id),
                "text": "something",
                "confidence": 0.369,
                "orientation": TextOrientation.HorizontalLeftToRight.value,
                "element": None,
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": self.worker_run.summary
                },
            }
        ])

    def test_list_transcriptions_recursive(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"recursive": "true"}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual(
            [
                (data["element"]["type"], data["element"]["zone"]["polygon"], data["text"])
                for data in response.json()["results"]
            ], [
                ("page", [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]], "Lorem ipsum dolor sit amet"),
                ("word", [[100, 100], [100, 200], [200, 200], [200, 100], [100, 100]], "PARIS"),
                ("word", [[400, 400], [400, 500], [500, 500], [500, 400], [400, 400]], "ROY"),
                ("word", [[700, 700], [700, 800], [800, 800], [800, 700], [700, 700]], "DATUM"),
            ]
        )

    def test_list_transcriptions_recursive_filter_element_type(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"recursive": "true", "element_type": "page"}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual(
            [
                (data["element"]["type"], data["element"]["zone"]["polygon"], data["text"])
                for data in response.json()["results"]
            ], [
                ("page", [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]], "Lorem ipsum dolor sit amet"),
            ]
        )

    def test_list_worker_version_transcriptions(self):
        worker_transcription = self.page.transcriptions.create(
            text="something",
            worker_version=self.worker_version_2,
            worker_run=self.worker_run,
            confidence=0.369,
        )

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"recursive": "true", "worker_version": str(self.worker_version_2.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json()["results"], [
            {
                "id": str(worker_transcription.id),
                "text": "something",
                "confidence": 0.369,
                "orientation": TextOrientation.HorizontalLeftToRight.value,
                "element": {
                    "id": str(self.page.id),
                    "name": "Volume 1, page 1r",
                    "type": "page",
                    "zone": {
                        "id": str(self.page.id),
                        "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    },
                    "rotation_angle": 0,
                    "mirrored": False
                },
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": self.worker_run.summary
                },
            }
        ])

    def test_list_manual_transcriptions(self):
        manual_transcription = self.page.transcriptions.create(
            text="something",
            confidence=0.369,
        )

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"recursive": "true", "worker_version": False}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json()["results"], [
            {
                "id": str(manual_transcription.id),
                "text": "something",
                "confidence": 0.369,
                "orientation": TextOrientation.HorizontalLeftToRight.value,
                "element": {
                    "id": str(self.page.id),
                    "name": "Volume 1, page 1r",
                    "type": "page",
                    "zone": {
                        "id": str(self.page.id),
                        "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    },
                    "rotation_angle": 0,
                    "mirrored": False
                },
                "worker_run": None,
            }
        ])

    def test_list_transcriptions_worker_version_validation(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"recursive": "true", "worker_version": "oh no"}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_version": ["“oh no” is not a valid UUID."]})

    def test_list_transcriptions_worker_run_validation(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"recursive": "true", "worker_run": "oh no"}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run": ["“oh no” is not a valid UUID."]})

    def test_list_worker_run_transcriptions_worker_version_filter(self):
        worker_transcription = self.page.transcriptions.create(
            text="something",
            worker_version=self.worker_version_2,
            worker_run=self.worker_run,
            confidence=0.369,
        )

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"recursive": "true", "worker_version": str(self.worker_version_2.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json()["results"], [
            {
                "id": str(worker_transcription.id),
                "text": "something",
                "confidence": 0.369,
                "orientation": TextOrientation.HorizontalLeftToRight.value,
                "element": {
                    "id": str(self.page.id),
                    "name": "Volume 1, page 1r",
                    "type": "page",
                    "zone": {
                        "id": str(self.page.id),
                        "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    },
                    "rotation_angle": 0,
                    "mirrored": False
                },
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": self.worker_run.summary
                },
            }
        ])

    def test_list_worker_run_transcriptions(self):
        worker_transcription = self.page.transcriptions.get(worker_run=self.worker_run)

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={"worker_run": str(self.worker_run.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json()["results"], [
            {
                "id": str(worker_transcription.id),
                "text": "Lorem ipsum dolor sit amet",
                "confidence": 1.0,
                "orientation": TextOrientation.HorizontalLeftToRight.value,
                "element": None,
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": "Worker Recognizer @ version 1",
                },
            }
        ])

    def test_list_worker_run_and_wrong_version(self):
        self.worker_version_1.transcriptions.update(worker_version=self.worker_version_2)

        self.client.force_login(self.user)

        # There should be one less query, but Django >=4.2 no longer ignores unions that are known to be empty
        # https://code.djangoproject.com/ticket/35559
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:element-transcriptions", kwargs={"pk": str(self.page.id)}),
                data={
                    "recursive": "true",
                    "worker_run": str(self.worker_run.id),
                    # Does not match the WorkerRun
                    "worker_version": str(self.worker_version_1.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json()["results"], [])

    # Elements listing with with_transcriptions param

    def test_list_elements_with_transcriptions_false(self):
        self.assertTrue(self.page.transcriptions.exists())
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(
                reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                data={"type": "page", "with_transcriptions": False}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 6)
        for elem in data["results"]:
            self.assertIsNone(elem["transcriptions"])

    def test_list_elements_with_transcriptions(self):
        tr1 = self.line.transcriptions.create(
            text="Demain dès l'aube, à l'heure où blanchit la campagne",
            worker_run=self.worker_run,
            worker_version=self.worker_run.version,
            confidence=0.89
        )
        tr2 = self.line.transcriptions.create(
            text="Je partirai, vois-tu je says que tu m'attends.",
            worker_run=self.worker_run,
            worker_version=self.worker_run.version,
            confidence=0.72
        )
        tl_2 = self.corpus.elements.create(
            type=self.line.type,
            name="Text Line 2",
            image=self.page.image,
            polygon=self.line.polygon
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(
                reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                data={"type": "text_line", "with_transcriptions": True}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 2)
        self.assertEqual(data["results"][0]["id"], str(self.line.id))
        self.assertListEqual(data["results"][0]["transcriptions"], [
            {
                "id": str(tr1.id),
                "text": "Demain dès l'aube, à l'heure où blanchit la campagne",
                "confidence": 0.89,
                "orientation": "horizontal-lr",
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": "Worker Recognizer @ version 1"
                }
            },
            {
                "id": str(tr2.id),
                "text": "Je partirai, vois-tu je says que tu m'attends.",
                "confidence": 0.72,
                "orientation": "horizontal-lr",
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": "Worker Recognizer @ version 1"
                }
            }
        ])
        self.assertEqual(data["results"][1]["id"], str(tl_2.id))
        self.assertListEqual(data["results"][1]["transcriptions"], [])

    def test_list_element_parents_with_transcriptions_false(self):
        self.assertTrue(self.page.transcriptions.exists())
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(
                reverse("api:elements-parents", kwargs={"pk": self.line.id}),
                data={"type": "page", "with_transcriptions": False}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], str(self.page.id))
        self.assertIsNone(data["results"][0]["transcriptions"])

    def test_list_element_parents_with_transcriptions(self):
        tr = self.page.transcriptions.get()
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(
                reverse("api:elements-parents", kwargs={"pk": self.line.id}),
                data={"type": "page", "with_transcriptions": True}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], str(self.page.id))
        self.assertListEqual(data["results"][0]["transcriptions"], [
            {
                "id": str(tr.id),
                "text": "Lorem ipsum dolor sit amet",
                "confidence": 1,
                "orientation": "horizontal-lr",
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": "Worker Recognizer @ version 1"
                }
            }
        ])

    def test_list_element_children_with_transcriptions_false(self):
        self.assertTrue(self.page.transcriptions.exists())
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": self.volume.id}),
                data={"type": "page", "with_transcriptions": False}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 3)
        self.assertEqual(data["results"][0]["id"], str(self.page.id))
        for page_elem in data["results"]:
            self.assertIsNone(page_elem["transcriptions"])

    def test_list_element_children_with_transcriptions(self):
        tr1 = self.line.transcriptions.create(
            text="Demain dès l'aube, à l'heure où blanchit la campagne",
            worker_run=self.worker_run,
            worker_version=self.worker_run.version,
            confidence=0.89
        )
        tr2 = self.line.transcriptions.create(
            text="Je partirai, vois-tu je says que tu m'attends.",
            worker_run=self.worker_run,
            worker_version=self.worker_run.version,
            confidence=0.72
        )
        tl_2 = self.corpus.elements.create(
            type=self.line.type,
            name="Text Line 2",
            image=self.page.image,
            polygon=self.line.polygon
        )
        tl_2.add_parent(self.page)

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": self.page.id}),
                data={"type": "text_line", "with_transcriptions": True}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 2)
        self.assertEqual(data["results"][0]["id"], str(self.line.id))
        self.assertListEqual(data["results"][0]["transcriptions"], [
            {
                "id": str(tr1.id),
                "text": "Demain dès l'aube, à l'heure où blanchit la campagne",
                "confidence": 0.89,
                "orientation": "horizontal-lr",
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": "Worker Recognizer @ version 1"
                }
            },
            {
                "id": str(tr2.id),
                "text": "Je partirai, vois-tu je says que tu m'attends.",
                "confidence": 0.72,
                "orientation": "horizontal-lr",
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": "Worker Recognizer @ version 1"
                }
            }
        ])
        self.assertEqual(data["results"][1]["id"], str(tl_2.id))
        self.assertListEqual(data["results"][1]["transcriptions"], [])
