import logging
import os
import urllib.parse
from json import JSONDecodeError

import requests
from django.conf import settings
from django.contrib.gis.db.models.functions import Scale, SnapToGrid
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import models
from django.db.models.functions import Concat, Substr
from django.utils.functional import cached_property
from django.utils.text import slugify
from enumfields import EnumField

from arkindex.images.managers import ImageServerManager
from arkindex.project.aws import S3FileMixin, S3FileStatus
from arkindex.project.fields import LStripTextField, MD5HashField, StripSlashURLField
from arkindex.project.models import IndexableModel
from arkindex.project.tools import should_verify_cert

logger = logging.getLogger(__name__)
profile_uri_validator = URLValidator(schemes=["http", "https"], message="Invalid IIIF profile URI")


class ImageServer(models.Model):
    """
    A IIIF image server
    """
    display_name = models.CharField(max_length=250)
    url = StripSlashURLField()

    s3_bucket = models.SlugField(max_length=63, blank=True, null=True, db_index=False)
    s3_region = models.SlugField(max_length=63, blank=True, null=True, db_index=False)
    s3_read_only_bucket = models.BooleanField(
        help_text="Disable the extra image checks that are normally made for servers with S3 buckets.",
        default=False,
    )

    max_width = models.PositiveIntegerField(blank=True, null=True)
    max_height = models.PositiveIntegerField(blank=True, null=True)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    objects = ImageServerManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                "url",
                name="unique_imageserver_url",
            ),
            # Enabling s3_read_only_bucket makes no sense when there is no s3_bucket,
            # so we force it to be disabled if there is None to avoid any confusions.
            models.CheckConstraint(
                condition=models.Q(s3_read_only_bucket=False) | models.Q(s3_bucket__isnull=False),
                name="s3_read_only_bucket_requires_s3_bucket"
            )
        ]

    def __str__(self):
        return self.display_name

    @property
    def is_local(self):
        """
        Indicates whether the server is the Arkindex IIIF server.
        """
        return self.id == settings.LOCAL_IMAGESERVER_ID

    def build_url(self, path):
        """
        Build a full url on this server
        """
        base_url = self.url
        path = path.lstrip("/")
        if not base_url.endswith("/"):
            base_url += "/"
        abs_url = urllib.parse.urljoin(base_url, path)

        # In some cases, the path will contain something that looks like a scheme,
        # e.g. ark:/something with Gallica URLs. In those cases, urljoin will not add
        # the server's root URL ; we fall back to a dirty way of joining the URLs.
        if abs_url.startswith(base_url):
            return abs_url

        return base_url + path  # Don't try this at home

    def merge_into(self, other):
        """
        Merge this ImageServer into another. This does not remove the ImageServer,
        it only moves around the images.
        This allows changing the root folder of a server and updating all of its images.
        Both servers must have the same URL schemes and domains.
        If both servers have the same root path, the images are simply moved.
        If the destination server is a parent folder of this server, the subfolder is prepended to all image paths.
        If the destination server is a child folder of this server, the subfolder is removed from all paths.
        In the latter case, merging is only allowed when all images of the server are in the same subfolder.
        Raises ValueError whenever merging is impossible.
        """
        assert isinstance(other, ImageServer)
        if not self.images.exists():
            return

        # Parse both URLs to first check for scheme/netloc compatibility
        src_url, dest_url = map(urllib.parse.urlsplit, (self.url, other.url))
        if src_url.scheme != dest_url.scheme:
            raise ValueError("Cannot merge into an image server of a different protocol")
        if src_url.netloc != dest_url.netloc:
            raise ValueError("Cannot merge into an image server on a different domain")

        # Check paths
        src_path, dest_path = src_url.path.strip("/"), dest_url.path.strip("/")
        common = os.path.commonpath([src_path, dest_path])

        if common == dest_path:
            # Destination is a parent folder of this server
            prefix = src_path[len(common):].lstrip("/")
            if not prefix.endswith("/"):
                prefix += "/"
            # Prepend all image paths with this new prefix and change their server
            self.images.update(
                server_id=other.id,
                path=Concat(models.Value(prefix), "path"),
            )
            return

        if common == src_path:
            # Destination is a child of this server
            prefix = dest_path[len(common):].lstrip("/")
            if not prefix.endswith("/"):
                prefix += "/"
            # Check that all images on this server start with the prefix
            if self.images.exclude(path__startswith=prefix).exists():
                raise ValueError("Cannot merge into an image server on a child folder: "
                                 "Some images are not in the child folder")
            # Remove the prefix from all image paths and change their server
            # Note that positions in Substr() are 1-indexed
            self.images.update(
                server_id=other.id,
                path=Substr("path", len(prefix) + 1),
            )
            return

        raise ValueError("Cannot merge into an image server on a different subfolder")

    def split(self, folder):
        """
        Create a new ImageServer for a subfolder on this server and move the images there.
        Raises ValueError if there are no images to move.
        Returns the newly created ImageServer.
        """
        # Ensure we have a path ending with a single slash
        folder_slash = folder.lstrip("/") + "/"
        images = self.images.filter(path__startswith=folder_slash)
        if not images.exists():
            raise ValueError(f'No images are in the "{folder}" folder')

        # Create the new server on the subfolder
        new_server = ImageServer.objects.create(
            display_name=f"{self.display_name}_{slugify(folder)}",
            url=urllib.parse.urljoin(self.url.lstrip("/") + "/", folder),
        )

        # Move the images there, removing the subfolder from their path
        images.update(
            server_id=new_server.id,
            # Substr positions are 1-indexed
            path=Substr("path", len(folder_slash) + 1)
        )
        return new_server


class Image(S3FileMixin, IndexableModel):
    """
    An element image
    """
    server = models.ForeignKey(ImageServer, on_delete=models.CASCADE, related_name="images")
    path = LStripTextField(chars="/")
    width = models.PositiveIntegerField(default=0)
    height = models.PositiveIntegerField(default=0)
    # Allow null values for external servers
    hash = MD5HashField(blank=True, null=True)
    status = EnumField(S3FileStatus, default=S3FileStatus.Unchecked, max_length=50)

    @cached_property
    def s3_bucket(self):
        return self.server.s3_bucket

    @cached_property
    def s3_region(self):
        return self.server.s3_region

    @cached_property
    def s3_key(self):
        return urllib.parse.unquote(self.path)

    class Meta:
        unique_together = (
            ("server", "path"),
            ("server", "hash"),
        )
        constraints = [
            # Require images to have non-zero dimensions when they have the Checked status
            models.CheckConstraint(
                condition=models.Q(width__gt=0, height__gt=0) | ~models.Q(status=S3FileStatus.Checked),
                name="checked_image_dimensions",
            ),
        ]

    @property
    def url(self) -> str:
        return self.server.build_url(self.path)

    def get_thumbnail_url(self, max_width=200, max_height=None):
        if max_width is None and max_height is None:
            resize = "full"
        else:
            resize = "{},{}".format(max_width or "", max_height or "")
        return "{}/full/{}/0/default.jpg".format(self.url.rstrip("/"), resize)

    @property
    def s3_put_url(self):
        assert self.server.is_local, "Cannot PUT images on remote servers"
        return super().s3_put_url

    def perform_check(self, save=True, raise_exc=False):
        """
        Check the image's existence and update width, height and status properties
        """
        url = self.url
        if not url.endswith("/"):
            url += "/"

        info_urls = set([urllib.parse.urljoin(url, "info.json"), f"{url}info.json"])
        for info_url in info_urls:
            requests_exception = None
            try:
                # Load info
                resp = requests.get(
                    info_url, timeout=15,
                    allow_redirects=True,
                    verify=should_verify_cert(info_url),
                    headers={"User-Agent": settings.IIIF_USER_AGENT}
                )
                resp.raise_for_status()
                try:
                    payload = resp.json()
                except JSONDecodeError:
                    raise ValueError("Server returned an invalid JSON document")
                self.check_from_payload(payload)
            except Exception as e:
                logger.warning(f"Image check failed on {info_url}: {e}")
                requests_exception = e
            else:
                break

        if requests_exception:
            self.status = S3FileStatus.Error
            if save:
                self.save()
            if raise_exc:
                raise requests_exception
        else:
            self.status = S3FileStatus.Checked
            if save:
                self.save()

    def check_from_payload(self, data):
        """
        Check and update the image's properties from an IIIF Image Information response
        This does not save in the DB nor handles any exceptions.
        """
        assert isinstance(data, dict), "Payload should be a dict"
        assert all(item in data for item in ("@id", "width", "height", "profile")), "Missing required properties"
        assert isinstance(data["profile"], (str, list)), "Profile is not a list or URI"

        # Check id
        image_id = data["@id"]
        assert image_id.startswith(self.server.url), \
            f"Image id does not start with server url ({image_id} vs. {self.server.url})"

        # Ensure that the required profile URI is a URI
        if isinstance(data["profile"], list):
            profile_uri = data["profile"][0]
        else:
            profile_uri = data["profile"]
        try:
            profile_uri_validator(profile_uri)
        except ValidationError as e:
            logger.warning(", ".join(list(e)))

        # Use image resource ID from IIIF server to update the image path if needed
        self.path = image_id[len(self.server.url) + 1:]

        new_width, new_height = int(data["width"]), int(data["height"])
        assert new_width > 0 and new_height > 0, "Image dimensions must be strictly positive integers"

        # Handle optional maxWidth and maxHeight properties in the profile description
        # In the IIIF Image API 2.x, `profile` is a list with one or more Image API compliance level URIs,
        # and optionally an object, the profile description.
        # In 1.x, `profile` is a single URI, not a list, and 1.x does not have maximum size properties.
        max_width, max_height = None, None
        if isinstance(data["profile"], list) and len(data["profile"]) >= 2:
            profile_desc = next((item for item in data["profile"] if isinstance(item, dict)), {})
            max_width = profile_desc.get("maxWidth")
            max_height = profile_desc.get("maxHeight")

        # Special case for Harvard IDS: those properties are wrongly defined in the main object
        # This is the proper location in IIIF 3.x, but Harvard provides IIIF 2 and we only support IIIF 2.
        if not max_width and "maxWidth" in data:
            max_width = data.get("maxWidth")
        if not max_height and "maxHeight" in data:
            max_height = data.get("maxHeight")

        # From the Image API specification: If maxWidth is specified and maxHeight is not,
        # then clients should infer that maxHeight = maxWidth.
        if max_width and not max_height:
            max_height = max_width

        # In some cases, the servers may return maximum widths/heights which are specific to an image,
        # and lower than the server-wide settings
        # Therefore, we can only warn when the server-wide settings are lower than what info.json states,
        # or when there are not defined.
        assert not any([
            (self.server.max_width is None) ^ (max_width is None),
            (self.server.max_height is None) ^ (max_height is None),
            self.server.max_width and max_width and self.server.max_width < max_width,
            self.server.max_height and max_height and self.server.max_height < max_height,
        ]), f"Image server maximum width/height ({self.server.max_width}×{self.server.max_height}) " \
            f"does not match the actual maximum width/height ({max_width}×{max_height}). " \
            "Please ask an instance administrator to update the IIIF server configuration."

        if self.width != new_width or self.height != new_height:
            # Check self.id to avoid resizing polygons on a non-existent image
            if self.id and self.width and self.height:
                ratio = min(new_width / self.width, new_height / self.height)

                # Update the element polygons to reflect updated image sizes
                if ratio != 1 and self.elements.exists():
                    self.elements.update(polygon=SnapToGrid(Scale(
                        models.F("polygon"),
                        x=ratio,
                        y=ratio,
                    ), 1))

            self.width, self.height = new_width, new_height

    def __str__(self):
        # Used by admin
        return self.url


class Thumbnail(S3FileMixin):
    """
    Describes an element thumbnail
    """
    s3_bucket = settings.AWS_THUMBNAIL_BUCKET

    def __init__(self, element):
        self.element = element

    @property
    def name(self):
        return f"{str(self.element.id.hex)}.jpg"

    @cached_property
    def s3_key(self):
        return self.name
