from django.conf import settings
from django.db.models import Count
from django.http import Http404, HttpResponse
from django.views import View

from arkindex.metrics.utils import build_metric
from arkindex.ponos.models import Agent, Farm, State, Task
from django_prometheus.exports import ExportToDjangoView


class PortFilterMixin:

    def dispatch(self, request, *args, **kwargs):
        if int(request.get_port()) != settings.PROMETHEUS_METRICS_PORT:
            raise Http404

        return super().dispatch(request, *args, **kwargs)


class MetricsView(PortFilterMixin, View):
    def get(self, request, *args, **kwargs):
        common_attributes = {
            "hostname": settings.PUBLIC_HOSTNAME,
            "env": settings.ARKINDEX_ENV
        }

        # Count nb of tasks in pending or unscheduled states
        # grouped by farm
        tasks_count = (
            Task.objects
            .filter(state__in=(State.Pending, State.Unscheduled))
            .values("process__farm", "state")
            .annotate(nb=Count("id"))
        )

        # Load all ponos agents details
        agents = Agent.objects.all().values("hostname", "farm_id", "last_ping")

        # Load all farms to use in attributes of farms & tasks
        farms = dict(Farm.objects.all().values_list("id", "name"))

        metrics = [
            # Backend ping
            build_metric(
                "arkindex_instance",
                attributes=common_attributes
            ),
        ] + [

            # Pending or Unscheduled tasks count
            build_metric(
                "arkindex_tasks",
                attributes={**common_attributes, "state": t["state"], "farm": farms.get(t["process__farm"])},
                value=t["nb"]
            )
            for t in tasks_count
        ] + [
            # Ponos agent last known ping
            build_metric(
                "arkindex_agent_ping",
                attributes={**common_attributes, "agent_name": a["hostname"], "farm": farms.get(a["farm_id"])},
                value=int(a["last_ping"].timestamp())
            )
            for a in agents
        ]

        # Render text response with all metrics
        return HttpResponse(
            "\n".join(metrics),
            content_type="text/plain"
        )


class DjangoPrometheusMetricsView(PortFilterMixin, View):

    def get(self, request, *args, **kwargs):
        # The django_prometheus view is a function, not a class-based view, so we have to build our own.
        return ExportToDjangoView(request)
