import uuid
from functools import partial
from textwrap import dedent

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.core.exceptions import ValidationError
from django.core.validators import MinLengthValidator, MinValueValidator
from django.db import models, transaction
from django.db.models import Exists, F, OuterRef, Q
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from enumfields import Enum, EnumField

import pgtrigger
from arkindex.documents.models import Classification, Element
from arkindex.ponos.models import FINAL_STATES, STATES_ORDERING, State, Task, token_default
from arkindex.process.managers import (
    ActivityManager,
    CorpusWorkerVersionManager,
    WorkerManager,
    WorkerRunManager,
    WorkerVersionManager,
)
from arkindex.project.aws import S3FileMixin, S3FileStatus
from arkindex.project.fields import ArrayField, MD5HashField
from arkindex.project.models import IndexableModel
from arkindex.project.tools import is_prefetched
from arkindex.project.validators import MaxValueValidator
from arkindex.training.models import ModelVersion, ModelVersionState
from arkindex.users.models import Role

ProcessBuilder = import_string(
    getattr(settings, "PROCESS_BUILDER_CLASS", None) or "arkindex.process.builder.ProcessBuilder"
)


def process_max_chunks():
    """
    Maximum value for Process.chunks.
    This must be defined as a named function because it will be serialized in Django migrations.
    """
    return settings.MAX_CHUNKS


def validate_unique_set_names(sets):
    if len(set(sets)) != len(sets):
        raise ValidationError("Set names must be unique.")


class ActivityState(Enum):
    """
    Store the state of the workers activity tracking for a process.
    To support large elements set, the state is asynchronously set to `ready` after a process
    has been started and worker activities have been initialized on its elements.
    """
    Disabled = "disabled"
    Pending = "pending"
    Ready = "ready"
    Error = "error"


class ProcessMode(Enum):
    Files = "files"
    Workers = "workers"
    Template = "template"
    S3 = "s3"
    Local = "local"
    Dataset = "dataset"
    Export = "export"


class ExportFormat(Enum):
    PDF = "pdf"
    PageXML = "page_xml"
    DOCX = "docx"
    CSV = "csv"


class Process(IndexableModel):

    name = models.CharField(null=True, blank=True, max_length=250)
    creator = models.ForeignKey("users.User", on_delete=models.CASCADE, related_name="processes")
    corpus = models.ForeignKey("documents.Corpus", on_delete=models.CASCADE, related_name="processes", null=True)
    mode = EnumField(ProcessMode, max_length=30)
    files = models.ManyToManyField("process.DataFile", related_name="processes")
    sets = models.ManyToManyField("training.DatasetSet", related_name="processes", through="process.ProcessDatasetSet")
    versions = models.ManyToManyField("process.WorkerVersion", through="process.WorkerRun", related_name="processes")
    activity_state = EnumField(ActivityState, max_length=32, default=ActivityState.Disabled)

    started = models.DateTimeField(blank=True, null=True)
    finished = models.DateTimeField(blank=True, null=True)

    farm = models.ForeignKey(
        to="ponos.Farm",
        related_name="processes",
        on_delete=models.DO_NOTHING,
        blank=True,
        null=True,
    )

    # Used to define the root element on which the process will run
    element = models.ForeignKey("documents.Element", on_delete=models.SET_NULL, null=True, blank=True)

    # Used to define the element type during folder elements creation for Images, PDF, IIIF and S3 imports
    folder_type = models.ForeignKey(
        "documents.ElementType",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="folder_imports",
    )

    # Used to define the element type during non-folder elements creation for Images, PDF, IIIF and S3 imports,
    # or to filter elements by type with Workers processes
    element_type = models.ForeignKey(
        "documents.ElementType",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="element_imports",
    )

    # Used to filter out elements with a name that doesn't contain the substring, only on Workers processes
    name_contains = models.CharField(null=True, blank=True, max_length=250)

    # Used to filter elements having a classification of a given MLClass in ListProcessElements
    ml_class = models.ForeignKey(
        "documents.MLClass",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="filtered_processes",
    )

    # Used to save a user's selection for Workers processes
    elements = models.ManyToManyField(
        "documents.Element",
        through="process.ProcessElement",
        related_name="processes",
    )

    # Used to include sub-elements of the actual query in Workers processes
    load_children = models.BooleanField(default=False)

    chunks = models.PositiveIntegerField(
        default=1,
        validators=[
            MinValueValidator(1),
            MaxValueValidator(process_max_chunks),
        ],
    )

    # Use elements cache in Workers processes
    use_cache = models.BooleanField(default=False)

    # Refers a process template
    template = models.ForeignKey(
        "process.Process",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="applications",
    )

    # S3 bucket name, and prefix within the bucket, for ProcessMode.S3
    # Bucket names are between 3 and 63 characters long
    bucket_name = models.CharField(
        max_length=63,
        blank=True,
        null=True,
        validators=[MinLengthValidator(3)],
    )
    # Object keys are up to 1024 characters long
    prefix = models.CharField(max_length=1024, blank=True, null=True)

    class Meta:
        ordering = ["corpus", "-created"]
        verbose_name_plural = "processes"
        constraints = [
            models.CheckConstraint(
                # Either the process mode is local and the process does not have a corpus,
                # or it has another mode and has a corpus set.
                condition=Q(mode=ProcessMode.Local) ^ ~Q(corpus=None),
                name="check_process_corpus",
                violation_error_message="Local processes cannot have a corpus, and other modes must have one set.",
            ),
            models.UniqueConstraint(
                "creator",
                condition=Q(mode=ProcessMode.Local),
                name="unique_local_process",
                violation_error_message="Only one local process is allowed per user.",
            ),
            models.CheckConstraint(
                condition=Q(started=None) | Q(started__gte=F("created")),
                name="process_started_after_created",
                violation_error_message="The process start date must not be earlier than the process creation date.",
            ),
            models.CheckConstraint(
                # This also implies that `started` must not be None if `finished` is set,
                # since finished >= NULL would be NULL and not TRUE.
                condition=Q(finished=None) | Q(finished__gte=F("started")),
                name="process_finished_after_started",
                violation_error_message="The process finish date must not be earlier than the process start date.",
            ),
        ]

    @property
    def state(self) -> State:
        """
        Deduce the process state from the tasks of its latest run.
        A process' state is deduced by picking the first state that any tasks in this run have, in the order defined by STATES_ORDERING.
        """
        return self.get_state(self.get_last_run())

    @property
    def is_final(self) -> bool:
        return self.state in FINAL_STATES

    @property
    def has_prefetched_tasks(self):
        """
        Returns whether or not all tasks of this process have been prefetched with .prefetch_related()
        or prefetch_related_objects().
        Prefetched tasks will only be accessible through `self.tasks.all()` and using any other method will cause another SQL query.
        Therefore, if this property is True, only .all() should be used, and Python methods should be used to filter or aggregate, instead of `.exists`, `.filter` or `.aggregate`.

        See https://stackoverflow.com/a/19651840/5990435
        """
        return is_prefetched(self.tasks)

    @property
    def expiry(self):
        """
        A process' expiry date. This is the latest expiry date of its tasks.
        No action is taken when a process is expired; the `cleanup` command alone uses it.
        """
        if self.has_prefetched_tasks:
            return max((t.expiry for t in self.tasks.all()), default=None)
        return self.tasks.aggregate(models.Max("expiry"))["expiry__max"]

    def get_last_run(self) -> int:
        """
        Get the last run number. If the `last_run` attribute is defined on this process,
        possibly from a `.annotate(last_run=Max('tasks__run'))` annotation in a Django QuerySet,
        this method will return the attribute's value instead of making another SQL query.
        """
        if not hasattr(self, "last_run"):
            # This prevents performing another SQL request when tasks have already been prefetched.
            # See https://stackoverflow.com/a/19651840/5990435
            if self.has_prefetched_tasks:
                self.last_run = max((t.run for t in self.tasks.all()), default=None)
            else:
                self.last_run = self.tasks.all().aggregate(models.Max("run"))["run__max"]

        if self.last_run is None:
            return -1

        return self.last_run

    def refresh_from_db(self, *args, **kwargs) -> None:
        # Remove the cached last_run if it was set, to ensure the process is properly refreshed
        if hasattr(self, "last_run"):
            del self.last_run
        super().refresh_from_db(*args, **kwargs)

    def get_state(self, run) -> State:
        """
        A separate method to get a process' state on a given run.

        Most users will only use the Process.state property to get the state for a process' last run.
        However, when trying to get a state for many processes at once, using ``.annotate(last_run=Max('tasks__run'))``
        and using the annotation with this method will prevent many useless SQL requests.

        Further performance improvements can be achieved with ``prefetch_related('tasks')``.
        """
        # Negative run numbers never have tasks
        if run < 0:
            return State.Unscheduled

        # This prevents performing another SQL request when tasks have already been prefetched.
        # See https://stackoverflow.com/a/19651840/5990435
        if self.has_prefetched_tasks:
            restarted_tasks = set(t.original_task_id for t in self.tasks.all())
            task_states = set(
                t.state
                for t in self.tasks.all()
                if t.run == run
                and t.id not in restarted_tasks
            )
        else:
            task_states = set(
                self.tasks
                .filter(run=run)
                # Skip tasks that have been restarted
                .exclude(Exists(Task.objects.filter(
                    process_id=OuterRef("process_id"),
                    run=OuterRef("run"),
                    original_task_id=OuterRef("pk"),
                )))
                .values_list("state", flat=True)
            )

        # This run has no tasks
        if not task_states:
            return State.Unscheduled

        # All tasks have the same state
        if len(task_states) == 1:
            return task_states.pop()

        for state in STATES_ORDERING:
            if state in task_states:
                return state

        raise NotImplementedError("Something went wrong")

    def _get_filters(self):
        filters = {
            "corpus_id": self.corpus_id,
        }
        if self.name_contains:
            filters["name__icontains"] = self.name_contains
        if self.element_type_id:
            filters["type_id"] = self.element_type_id
        if self.ml_class_id:
            filters["id__in"] = Classification.objects.filter(ml_class_id=self.ml_class_id).values("element_id")

        return filters

    def copy_runs(self, new_process):
        """
        Copies this process' WorkerRuns to another process.

        This uses a dict to remap the existing parent UUIDs to the newly created UUIDs:
        otherwise, the runs on the other process would still have as their parents
        the runs of the current process, breaking their dependencies.
        """
        # Keep the relation between template_worker_runs and new_process_worker_runs ids
        new_runs = {}

        # First we copy each worker runs, excluding elements initialisation worker runs
        for run in self.worker_runs.all():
            if run.version.is_init_elements():
                continue

            # Create a new WorkerRun with same version, configuration and parents.
            new_run = WorkerRun(
                process=new_process,
                version_id=run.version_id,
                model_version_id=run.model_version_id,
                parents=run.parents,
                configuration_id=run.configuration_id,
                # Force use_gpu to match the GPU usage on WorkerVersions, so that invalid values
                # from the original process do not contaminate the new processes.
                use_gpu=(
                    run.use_gpu
                    if run.version.gpu_usage == FeatureUsage.Supported
                    else run.version.gpu_usage == FeatureUsage.Required
                ),
                summary=run.summary,
            )
            # Save the correspondence between this process' worker_run and the new one
            new_runs[run.id] = new_run

        # Remap parent ids correctly
        for run in new_runs.values():
            # If there are parents, we need to use the correct worker_run id, the newly created one
            run.parents = [
                new_runs[parent_id].id for parent_id in run.parents
                if parent_id in new_runs
            ]
        WorkerRun.objects.bulk_create(new_runs.values())

    def list_elements(self):
        """
        Return a queryset of elements involved in this process
        """
        if self.mode not in [ProcessMode.Workers, ProcessMode.Export]:
            return Element.objects.none()

        elements = None

        if self.element:
            # Assert the element has the same corpus as the process
            assert self.element.corpus_id == self.corpus_id, \
                f"Element {self.element.id} is not part of corpus {self.corpus_id}"
            # Handle a single element
            elements = Element.objects.filter(id=self.element.id)

        elif self.elements.exists():
            # Handle a selection of elements
            elements = self.elements.all()
            # Check all elements are in the same corpus as the process
            corpus = list(elements.values_list("corpus_id", flat=True).distinct())
            assert len(corpus) == 1 and corpus[0] == self.corpus_id, \
                f"Some elements on this process are not part of corpus {self.corpus_id}"

        if elements is not None and self.load_children:
            # Load all the children elements whose path contains the pre-selected elements
            # Those children are appended to the pre-selection
            elements |= Element.objects.filter(
                paths__path__overlap=map(str, elements.values_list("id", flat=True)),
            )

        # Load the full corpus, only when elements has not been populated before
        if elements is None:
            # Handle all elements of the process' corpus
            elements = Element.objects.filter(corpus=self.corpus_id)
            if not self.load_children:
                # Use top-level elements only
                elements = elements.filter(paths__path=[])

        # Filter elements depending on process properties
        return elements.filter(**self._get_filters())

    @transaction.atomic
    def run(self):
        """
        Build and start a new run for this process.
        """
        from arkindex.project.triggers import schedule_tasks

        process_builder = ProcessBuilder(self)
        process_builder.validate()
        process_builder.build()
        # Save all tasks and their relations
        process_builder.save()
        if settings.PONOS_RQ_EXECUTION:
            # Trigger tasks execution in RQ after the current transaction so tasks have been created
            transaction.on_commit(partial(schedule_tasks, process=self, run=process_builder.run))

        self.started = timezone.now()
        self.finished = None

        last_run = self.get_last_run()

        # Asynchronously initialize worker activities on Workers processes
        # On the very first run, we will generate them only if they are pending;
        # maybe something caused the activities to be prepared before running the process.
        # On subsequent runs, we always reinitialize.
        # We always have worker activities on Workers processes, so disabled is the same as pending.
        if (
            self.mode == ProcessMode.Workers
            and (self.activity_state in (ActivityState.Disabled, ActivityState.Pending) or last_run >= 0)
        ):
            from arkindex.project.triggers import initialize_activity
            initialize_activity(self)
            self.activity_state = ActivityState.Pending

        self.save()

        # Save an SQL query when re-serializing this process, since this new run is the last run
        self.last_run = last_run + 1

    def stop(self) -> None:
        """
        Fully stop the workflow by updating every running task to the Stopping state,
        and every unscheduled task to the Stopped state.
        """
        assert not self.is_final, "Only processes in a final state can be stopped"

        stopping_count = self.tasks.filter(
            state__in=[State.Pending, State.Running]
        ).update(state=State.Stopping)
        self.tasks.filter(state=State.Unscheduled).update(state=State.Stopped)

        # If all the tasks are immediately stopped, then UpdateTask will not be able to update
        # the finished attribute, so we do it here.
        if not stopping_count:
            self.finished = timezone.now()
            self.save()

    def clear(self):
        self.worker_runs.all().delete()
        self.template_id = None
        self.save()


class ProcessElement(models.Model):
    """
    Link between Processes and Elements to allow workflows from selections.
    May be removed later once the ElementPath algorithms are optimized,
    as a selection could then be turned into a set.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey(Process, on_delete=models.CASCADE, related_name="process_elements")
    element = models.ForeignKey("documents.Element", on_delete=models.CASCADE, related_name="process_elements")

    class Meta:
        unique_together = (
            ("process", "element"),
        )


class ProcessDatasetSet(models.Model):
    """
    Link between Processes and Dataset Sets.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey(Process, on_delete=models.CASCADE, related_name="process_sets")
    set = models.ForeignKey("training.DatasetSet", on_delete=models.DO_NOTHING, related_name="process_sets")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["process", "set"],
                name="unique_process_set",
            )
        ]


class DataFile(S3FileMixin, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    size = models.PositiveIntegerField(help_text="file size in bytes")
    content_type = models.CharField(max_length=120)
    corpus = models.ForeignKey("documents.Corpus", on_delete=models.CASCADE, related_name="files")
    status = EnumField(S3FileStatus, default=S3FileStatus.Unchecked, max_length=50)
    trashed = models.BooleanField(default=False)

    s3_bucket = settings.AWS_STAGING_BUCKET

    @cached_property
    def s3_key(self):
        return str(self.id)

    class Meta:
        ordering = ["corpus", "name"]

    def perform_check(self, save=True, raise_exc=False):
        """
        Check Datafile's existence, content type and size from S3
        """
        try:
            assert self.exists(), "File has not been uploaded"
            assert self.s3_object.content_length == self.size, \
                f"Uploaded file size is {self.s3_object.content_length} bytes, expected {self.size} bytes"
            assert self.s3_object.content_type == self.content_type, \
                f"Uploaded file content type is {self.s3_object.content_type}, expected {self.content_type}"
            self.status = S3FileStatus.Checked
        except AssertionError:
            self.status = S3FileStatus.Error
            if raise_exc:
                raise
        finally:
            if save:
                self.save()


class Worker(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    slug = models.CharField(max_length=100)
    type = models.ForeignKey("process.WorkerType", on_delete=models.PROTECT, related_name="workers")
    description = models.TextField(default="", blank=True)
    memberships = GenericRelation("users.Right", "content_id")

    # Allow any authentified user to see and execute this worker
    public = models.BooleanField(default=False)

    archived = models.DateTimeField(null=True, blank=True, default=None)

    repository_url = models.URLField(null=True, blank=True, max_length=250, default=None)

    objects = WorkerManager()

    # Worker costs
    cost_cpu_hour = models.DecimalField(
        max_digits=8,
        decimal_places=3,
        default=0,
        validators=[MinValueValidator(0)],
        verbose_name="CPU cost per hour",
        help_text=dedent("""
            Cost for running any WorkerVersion of this worker for an hour with `use_gpu = False` set on the WorkerRun, in euros.

            Irrelevant if the WorkerVersion requires GPUs.
        """).strip(),
    )
    cost_gpu_hour = models.DecimalField(
        max_digits=8,
        decimal_places=3,
        default=0,
        validators=[MinValueValidator(0)],
        verbose_name="GPU cost per hour",
        help_text=dedent("""
            Cost for running any WorkerVersion of this worker for an hour with `use_gpu = True` set on the WorkerRun, in euros.

            Irrelevant if the WorkerVersion does not support GPUs.
        """).strip(),
    )
    cost_1k_elements = models.DecimalField(
        max_digits=8,
        decimal_places=3,
        default=0,
        validators=[MinValueValidator(0)],
        verbose_name="Cost per 1000 elements",
        help_text="Cost for running any WorkerVersion of this worker over 1000 elements, in euros.",
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["slug", "repository_url"],
                name="worker_unique_slug_per_repository_url",
                # Treat null values the same as other values, as by default null values would allow duplicates
                # This allows slugs to be unique on local workers without repository URLs
                nulls_distinct=False,
            ),
        ]

    def __str__(self):
        return f"{self.name}"

    def is_readable(self, user) -> bool:
        """
        Whether a user can view this worker
        """
        # Require authentication, block viewing by Ponos agents (not Ponos tasks)
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False
        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self)
        return self.public or (level is not None and level >= Role.Contributor.value)

    def is_editable(self, user) -> bool:
        """
        Whether a user can edit this worker
        """
        # Require authentication, block edition by Ponos agents (not Ponos tasks)
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False

        # Allow admins to edit anything
        if user.is_admin:
            return True

        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self)

        # Otherwise, users should have contributor access to the worker
        return level is not None and level >= Role.Contributor.value

    def is_executable(self, user) -> bool:
        """
        Whether a user can execute this worker
        """
        # Require authentication, block execution by Ponos agents (not Ponos tasks)
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False

        # Apply the same rules as for editable workers,
        # but also allow any authenticated user to execute a public worker
        return self.public or self.is_editable(user)

    def is_archivable(self, user) -> bool:
        """
        Whether the user can archive or unarchive this worker
        """
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False

        if user.is_admin:
            return True

        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self)

        return level is not None and level >= Role.Admin.value

    def is_admin(self, user) -> bool:
        """
        Whether or not the user has administrator rights on this worker
        """
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False

        if user.is_admin:
            return True

        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self)

        return level is not None and level >= Role.Admin.value


class WorkerType(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = models.SlugField(max_length=100, db_index=False)
    display_name = models.CharField(max_length=100)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                "slug",
                name="unique_workertype_slug",
            ),
        ]

    def __str__(self):
        return self.slug

    def __repr__(self):
        return f"<WorkerType {self.slug}>"


class WorkerVersionState(Enum):
    Created = "created"
    Available = "available"
    Error = "error"


class FeatureUsage(Enum):
    Disabled = "disabled"
    Supported = "supported"
    Required = "required"


class ArkindexFeature(Enum):
    InitElements = "init_elements"
    FileImport = "file_import"
    S3Ingest = "s3_ingest"
    DatasetExtractor = "dataset_extractor"
    # When adding a new export worker, don't forget to also update the ExportFormat enum and the
    # FEATURE_FORMAT_MAP dictionary which maps export formats to arkindex features
    ExportPDF = "pdf_export"
    ExportPageXML = "pagexml_export"
    ExportDOCX = "docx_export"
    ExportCSV = "csv_export"


class WorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    worker = models.ForeignKey("process.Worker", on_delete=models.CASCADE, related_name="versions")
    version = models.PositiveIntegerField(
        validators=[MinValueValidator(1)],
        null=True,
        blank=True,
    )
    configuration = models.JSONField(blank=True, default=dict)
    state = EnumField(WorkerVersionState, default=WorkerVersionState.Created)
    gpu_usage = EnumField(FeatureUsage, default=FeatureUsage.Disabled, blank=True)
    model_usage = EnumField(FeatureUsage, default=FeatureUsage.Disabled, blank=True)

    # The Docker internal image id (sha256:xxx) that can be shared across multiple images
    # For workers imported with `arkindex workers publish`, the tag of the Docker image
    docker_image_iid = models.CharField(null=True, blank=True, max_length=512)

    # URL of the commit for this version, when worker is based on a repository
    revision_url = models.URLField(null=True, blank=True, max_length=250, default=None)
    branch = models.CharField(blank=True, null=True, max_length=512, default=None)
    tag = models.CharField(blank=True, null=True, max_length=512, default=None, db_collation="english_numeric")

    feature = EnumField(
        ArkindexFeature,
        max_length=20,
        blank=True,
        null=True,
        help_text="An Arkindex feature that this worker version provides. "
                  "There can only be one worker version per feature across an Arkindex instance.",
    )

    corpora = models.ManyToManyField(
        "documents.Corpus",
        through="process.CorpusWorkerVersion",
        related_name="worker_versions",
    )
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    objects = WorkerVersionManager()

    class Meta:
        constraints = [
            # Available worker versions must have a docker_image_iid set
            models.CheckConstraint(
                condition=~models.Q(state=WorkerVersionState.Available) | ~Q(docker_image_iid=None),
                name="workerversion_available_requires_docker_image",
            ),
            models.CheckConstraint(
                condition=Q(revision_url__isnull=True) ^ Q(version__isnull=True),
                name="workerversion_revision_url_xor_version",
            ),
            models.CheckConstraint(
                condition=models.Q(configuration__typeof="object"),
                name="worker_version_configuration_objects",
            ),
            models.UniqueConstraint(
                fields=["worker", "revision_url"],
                name="workerversion_unique_revision_url",
                condition=Q(revision_url__isnull=False),
            ),
            models.UniqueConstraint(
                fields=["worker", "version"],
                name="workerversion_unique_version",
                condition=Q(version__isnull=False),
            ),
            models.UniqueConstraint(
                fields=["worker", "tag"],
                name="workerversion_unique_tag",
                condition=Q(tag__isnull=False),
            ),
            models.UniqueConstraint(
                fields=["worker", "branch"],
                name="workerversion_unique_branch",
                condition=Q(branch__isnull=False),
            ),
            models.UniqueConstraint(
                fields=["feature"],
                name="workerversion_unique_feature",
                condition=Q(feature__isnull=False),
                violation_error_message="There can only be one worker version per feature across an Arkindex instance.",
            ),
            models.CheckConstraint(
                condition=Q(state=WorkerVersionState.Available) | Q(feature=None),
                name="workerversion_feature_requires_available",
                violation_error_message="A worker version can only provide an Arkindex feature if it is marked as available.",
            ),
        ]

    def __str__(self):
        if self.revision_url is None:
            return f"{self.worker} for version {self.version}"
        return f"{self.worker} for revision {self.revision_url}"

    @property
    def slug(self):
        return f"{self.worker.slug}_{self.truncated_id}"

    @property
    def truncated_id(self):
        return str(self.id)[0:6]

    @property
    def docker_command(self):
        if "docker" not in self.configuration:
            return
        return self.configuration["docker"].get("command")

    @property
    def docker_shm_size(self):
        if "docker" not in self.configuration:
            return
        return self.configuration["docker"].get("shm_size")

    @property
    def required_user_configuration_fields(self):
        """
        Set of field names defined in this worker version's user configuration
        which are required and without a default value.

        A worker configuration will need to be set on the worker run with those fields set
        before this worker version can be executed.
        """
        return set(
            name
            for name, options in self.configuration.get("user_configuration", {}).items()
            # Only pick the fields that are required and without default values
            if options.get("required") and "default" not in options
        )

    def is_init_elements(self):
        return (
            self.feature == ArkindexFeature.InitElements
            # Make detection of the init task in case its version changes
            or self.worker.slug.startswith("init-elements")
            or self.worker.slug.startswith("init_elements")
            or self.worker.slug.startswith("initialisation")
        )



class WorkerConfiguration(IndexableModel):
    name = models.CharField(max_length=250, validators=[MinLengthValidator(1)])
    configuration = models.JSONField(default=dict)
    configuration_hash = MD5HashField()
    archived = models.BooleanField(default=False)
    worker = models.ForeignKey(
        Worker,
        on_delete=models.CASCADE,
        related_name="configurations",
    )

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<WorkerConfiguration {self.name!r} ({self.id})>"

    class Meta:
        unique_together = (
            ("worker", "configuration_hash"),
            ("worker", "name")
        )
        constraints = [
            models.CheckConstraint(
                condition=models.Q(configuration__typeof="object"),
                name="worker_configuration_configuration_objects",
            ),
            models.CheckConstraint(
                condition=~Q(name=""),
                name="worker_configuration_name_not_empty",
                violation_error_message="Worker configuration name cannot be empty.",
            ),
        ]


class WorkerRun(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey("process.Process", on_delete=models.CASCADE, related_name="worker_runs")
    version = models.ForeignKey("process.WorkerVersion", on_delete=models.CASCADE, related_name="worker_runs")
    model_version = models.ForeignKey("training.ModelVersion", on_delete=models.CASCADE, related_name="worker_runs", blank=True, null=True)
    parents = ArrayField(models.UUIDField(), blank=True, default=list)
    configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name="worker_runs",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    summary = models.TextField()
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    has_results = models.BooleanField(default=False)
    use_gpu = models.BooleanField(default=False)

    objects = WorkerRunManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["process", "version", "model_version", "configuration"],
                name="unique_workerrun",
                # Treat null values the same as other values, as by default null values would allow duplicates
                nulls_distinct=False,
            ),
        ]

    @property
    def task_slug(self):
        """
        A slug that can be used to create a Ponos task from this WorkerRun.
        This does not include the chunk suffix when chunks are enabled.
        """
        # Since WorkerVersion can be used multiple times in a process,
        # we cannot depend only on the WorkerVersion to make a unique task slug:
        # we add the WorkerRun ID at the end of the slug
        return f"{self.version.worker.slug}_{str(self.id)[:6]}"

    @property
    def has_costs(self) -> bool:
        """
        Whether running a task created from this WorkerRun could incur costs in a budget
        """
        # Use either the CPU or GPU cost depending on how we are running this worker
        hourly_cost = self.version.worker.cost_gpu_hour if self.use_gpu else self.version.worker.cost_cpu_hour
        # Only apply the cost per element when the process might have worker activities
        return hourly_cost > 0 or (self.process.activity_state != ActivityState.Disabled and self.version.worker.cost_1k_elements > 0)

    def build_task(self, process, env, import_task_name, elements_path, run=0, chunk=None, workflow_runs=None, active_gpu_agents=False):
        """
        Build the Task that will represent this WorkerRun in ponos using :
        - the docker image name given by the WorkerVersion
        - the command given by the WorkerVersion
        - the artifact given by the WorkerVersion that will help us download the distant Docker image (S3)
        Return that Task and the list of its parent tasks slugs, to be added after they have all been created.
        """
        slug_suffix = f"_{chunk}" if chunk else ""

        parents = None
        if self.parents:
            if workflow_runs:
                # Use already fetched worker runs if possible
                parent_runs = [run for run in workflow_runs if run.id in self.parents]
            else:
                parent_runs = (
                    WorkerRun
                    .objects
                    .using("default")
                    .filter(id__in=self.parents)
                    .select_related("version__worker")
                    # The IDs are required so that Django understands that everything is prefetched.
                    # WorkerVersion.slug will just require the version ID and worker slug
                    # The init_elements checks requires the version's feature field
                    .only("id", "version_id", "version__worker_id", "version__worker__slug", "version__feature")
                )
            parents = [
                worker_run.task_slug + slug_suffix
                # If the parent is the initialisation task, no suffix is added as it is not being split in
                # chunks and the import_task_name is used, not worker_run.task_slug
                if not worker_run.version.is_init_elements()
                else import_task_name
                for worker_run in parent_runs
            ]
        elif import_task_name:
            parents = [import_task_name]
        else:
            parents = []

        assert (
            self.version.state == WorkerVersionState.Available
            and self.version.docker_image_iid
        ), (
            f"Worker Version {self.version.id} is not available and cannot be used to build a task."
        )

        task_env = env.copy()
        token = token_default()
        task_env["ARKINDEX_TASK_TOKEN"] = token
        task_env["TASK_ELEMENTS"] = elements_path
        task_env["ARKINDEX_WORKER_RUN_ID"] = str(self.id)
        if chunk:
            task_env["ARKINDEX_TASK_CHUNK"] = str(chunk)

        extra_files = {}
        # If the worker run needs a model_version, provide its public download url as extra file
        if self.model_version:
            assert self.model_version.state == ModelVersionState.Available, f"ModelVersion {self.model_version.id} is not available and cannot be used to build a task."
            extra_files = {"model": settings.PUBLIC_HOSTNAME + reverse("api:model-version-download", kwargs={"pk": self.model_version.id}) + f"?token={self.model_version.build_authentication_token_hash()}"}

        requires_gpu = self.use_gpu and self.version.gpu_usage in (FeatureUsage.Required, FeatureUsage.Supported)

        task = Task(
            command=self.version.docker_command,
            image=self.version.docker_image_iid,
            env=task_env,
            shm_size=self.version.docker_shm_size,
            slug=self.task_slug + slug_suffix,
            # The depth will be recomputed before creating all tasks for the process
            depth=0,
            run=run,
            token=token,
            process=process,
            worker_run=self,
            extra_files=extra_files,
            requires_gpu=requires_gpu,
            ttl=process.corpus.applied_maximum_task_ttl,
        )

        return task, parents

    def build_summary(self):
        """
        Build a summary of the worker run displaying information about :
        - worker version used
        - model version used if provided
        - worker configuration loaded if there is one
        """

        summary_text = f"Worker {self.version.worker.name} @ "
        if self.version.tag is not None or self.version.branch is not None:
            if self.version.tag is not None:
                summary_text += f"{self.version.tag} "
            else:
                summary_text += f"{self.version.branch} "

            if self.version.revision_url is not None:
                summary_text += f"({self.version.truncated_id})"
            else:
                summary_text += f"(version {self.version.version})"
        else:
            if self.version.revision_url is not None:
                summary_text += self.version.truncated_id
            else:
                summary_text += f"version {self.version.version}"

        if self.model_version:
            summary_text += f" with model {self.model_version.model.name} @ {self.model_version.truncated_id}"

        if self.configuration:
            summary_text += f" using configuration '{self.configuration.name}'"

        self.summary = summary_text


class WorkerActivityState(Enum):
    Queued = "queued"
    Started = "started"
    Processed = "processed"
    Error = "error"


class WorkerActivity(models.Model):
    """
    Many-to-many relationship between Element and WorkerVersion
    Used to track the activity of a worker version among multiple elements
    """
    # Using an UUID helps to execute SQL raw INSERT
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    created = models.DateTimeField(auto_now_add=True)

    # We will manually handle the timestamp update using a PostgreSQL trigger to avoid issues with bulk inserts/updates,
    # and we still let Django handle setting the timestamp initially, so we use `auto_now_add` and not `auto_now`.
    updated = models.DateTimeField(auto_now_add=True)

    # To properly compute a processing time, we need to separately store the time at which the state was updated to `started`.
    started = models.DateTimeField(blank=True, null=True)

    element = models.ForeignKey(
        "documents.Element",
        on_delete=models.CASCADE,
        related_name="activities",
    )
    worker_version = models.ForeignKey(
        WorkerVersion,
        on_delete=models.CASCADE,
        related_name="activities",
    )
    state = EnumField(
        WorkerActivityState,
        default=WorkerActivityState.Queued,
    )
    process = models.ForeignKey(
        Process,
        related_name="activities",
        on_delete=models.CASCADE,
    )
    configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name="worker_activities",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    model_version = models.ForeignKey(
        ModelVersion,
        related_name="worker_activities",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    # Specific WorkerActivity manager
    objects = ActivityManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["worker_version", "element", "configuration", "model_version"],
                name="workeractivity_unique",
                # Treat null values the same as other values, as by default null values would allow duplicates
                nulls_distinct=False,
            ),
            # Require the `started` timestamp to be set if a WorkerActivity is in a `started` state.
            # This does not prevent activities in a processed/error state to not have start times,
            # as we do not know when activities that existed before the start time was added have started,
            # but is enough to ensure that we have to set the start time on any activity that now becomes started.
            models.CheckConstraint(
                condition=~Q(state=WorkerActivityState.Started) | Q(started__isnull=False),
                name="worker_activity_started_requires_started",
            ),
        ]
        triggers = [
            pgtrigger.Trigger(
                name="update_workeractivity_updated",
                operation=pgtrigger.Update,
                when=pgtrigger.Before,
                func="NEW.updated = now(); RETURN NEW;",
            ),
            pgtrigger.ReadOnly(
                name="read_only_workeractivity_updated",
                fields=["updated"],
            )
        ]


class CorpusWorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    corpus = models.ForeignKey(
        "documents.Corpus",
        on_delete=models.CASCADE,
        related_name="worker_version_cache",
    )
    worker_version = models.ForeignKey(
        WorkerVersion,
        on_delete=models.CASCADE,
        related_name="corpus_cache",
    )
    model_version = models.ForeignKey(
        "training.ModelVersion",
        related_name="corpus_worker_versions",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    worker_configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name="corpus_worker_versions",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    objects = CorpusWorkerVersionManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["corpus", "worker_version", "model_version", "worker_configuration"],
                name="corpus_workerversion_unique",
                # Treat null values the same as other values, as by default null values would allow duplicates
                nulls_distinct=False,
            ),
        ]


# A mapping of export formats with their corresponding arkindex feature
FEATURE_FORMAT_MAP = {
    ExportFormat.PDF: ArkindexFeature.ExportPDF,
    ExportFormat.PageXML: ArkindexFeature.ExportPageXML,
    ExportFormat.DOCX: ArkindexFeature.ExportDOCX,
    ExportFormat.CSV: ArkindexFeature.ExportCSV
}
