from collections import defaultdict
from enum import Enum
from textwrap import dedent

from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import transaction
from django.db.models import Max, Q
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.ponos.models import Task
from arkindex.process.models import (
    CorpusWorkerVersion,
    FeatureUsage,
    Process,
    Worker,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.process.utils import hash_object
from arkindex.project.serializer_fields import ArchivedField, EnumField
from arkindex.training.models import Model
from arkindex.training.serializers import ModelVersionLightSerializer
from arkindex.users.models import Role
from docker.auth import resolve_repository_name
from docker.errors import InvalidRepository
from docker.utils import parse_repository_tag
from packaging.version import InvalidVersion
from packaging.version import parse as parse_version


class ExistingWorkerException(Exception):
    def __init__(self, worker):
        self.worker = worker


class WorkerSerializer(serializers.ModelSerializer):
    archived = ArchivedField(required=False)
    type = serializers.SlugField(
        help_text=dedent("""
            Slug of a WorkerType to set on this worker.

            If the WorkerType does not exist, it is created automatically.
        """)
    )

    class Meta:
        model = Worker
        fields = ("id", "name", "description", "type", "slug", "repository_url", "archived", "cost_cpu_hour", "cost_gpu_hour", "cost_1k_elements")
        read_only_fields = ("id", "cost_cpu_hour", "cost_gpu_hour", "cost_1k_elements")
        extra_kwargs = {
            "description": {
                "style": {"base_template": "textarea.html"},
            },
            # Type can be omitted when updating a worker
            "type": {
                "required": False
            }
        }
        validators = [
            serializers.UniqueTogetherValidator(
                queryset=Worker.objects.all(),
                fields=["repository_url", "slug"],
                message="A worker with the same repository URL and slug already exists.",
            ),
        ]

    def update(self, instance, validated_data):
        type_slug = validated_data.pop("type", None)
        if type_slug and type_slug != instance.type.slug:
            worker_type, _ = WorkerType.objects.using("default").get_or_create(
                slug=type_slug,
                defaults={"display_name": type_slug.capitalize()}
            )
            validated_data["type_id"] = worker_type.id
        return super().update(instance, validated_data)


class WorkerCreateSerializer(WorkerSerializer):
    repository_url = serializers.URLField(allow_null=True, required=False)

    # Creating an archived worker makes no sense
    archived = ArchivedField(read_only=True)

    class Meta(WorkerSerializer.Meta):
        # Remove the validator for unique slugs/repo URL since we can return an existing worker
        validators = []

    def create(self, validated_data):
        """
        Retrieve or create the worker and its type.
        """
        request = self.context["request"]
        with transaction.atomic(using="default"):
            # First check no worker exists with the given slug and repository URL
            existing_worker = (
                Worker.objects
                .using("default")
                .filter(
                    repository_url=validated_data.get("repository_url"),
                    slug=validated_data["slug"],
                )
                .first()
            )
            if existing_worker:
                if not existing_worker.is_executable(user=request.user):
                    raise ValidationError({
                        "repository_url": ["You cannot use this value."],
                        "slug": ["You cannot use this value."],
                    })
                raise ExistingWorkerException(existing_worker)

            # Get or create the worker and its type
            worker_type, _ = WorkerType.objects.using("default").get_or_create(
                slug=validated_data["type"],
                defaults={
                    "display_name": validated_data["type"].capitalize()
                }
            )
            worker = Worker.objects.create(
                slug=validated_data["slug"],
                name=validated_data["name"],
                type=worker_type,
                repository_url=validated_data.get("repository_url")
            )
            # Automatically grant an admin access to the creator of a local worker
            worker.memberships.create(user=request.user, level=Role.Admin.value)
            return worker


class WorkerTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = WorkerType
        fields = (
            "id",
            "slug",
            "display_name"
        )


class UserConfigurationFieldType(Enum):
    Int = "int"
    Float = "float"
    String = "string"
    Enum = "enum"
    Boolean = "bool"
    Dict = "dict"
    List = "list"
    Model = "model"


USER_CONFIGURATION_FIELD_TYPE_DATA_TYPE = {
    UserConfigurationFieldType.Int: [serializers.IntegerField(), int],
    UserConfigurationFieldType.Float: [serializers.FloatField(), (int, float)],
    UserConfigurationFieldType.String: [serializers.CharField(), str],
    UserConfigurationFieldType.Enum: [serializers.ChoiceField(choices=[]), None],
    UserConfigurationFieldType.Boolean: [serializers.BooleanField(), bool],
    UserConfigurationFieldType.Dict: [serializers.DictField(child=serializers.CharField()), dict],
    UserConfigurationFieldType.List: [serializers.ListField(), list],
    UserConfigurationFieldType.Model: [serializers.UUIDField(), str]
}


class UserConfigurationFieldSerializer(serializers.Serializer):
    title = serializers.CharField()
    type = EnumField(UserConfigurationFieldType)
    subtype = EnumField(UserConfigurationFieldType, required=False)
    required = serializers.BooleanField(default=False)
    choices = serializers.ListField(required=False, allow_empty=False, allow_null=True)
    multiline = serializers.BooleanField(required=False)

    def to_internal_value(self, data):
        errors = defaultdict(list)

        allowed_fields = ("title", "type", "required", "default", "choices", "subtype", "multiline")

        if not isinstance(data, dict):
            errors["__all__"] = [f"User configuration field definitions should be of type dict, not {type(data).__name__}."]
            raise ValidationError(errors)

        for field in data:
            if field not in allowed_fields:
                errors[field].append(
                    f"Configurable properties can only be defined using the following keys: {', '.join(allowed_fields)}."
                )

        default_value = data.get("default")
        data = super().to_internal_value(data)

        field_type = data.get("type")
        choices = data.get("choices")
        subtype = data.get("subtype")

        if field_type == UserConfigurationFieldType.List and not subtype:
            errors["subtype"].append('The "subtype" field must be set for "list" type properties.')

        # Handle subtypes
        if subtype is not None:
            if field_type != UserConfigurationFieldType.List:
                errors["subtype"].append('The "subtype" field can only be set for a "list" type property.')
            if subtype not in [UserConfigurationFieldType.Int, UserConfigurationFieldType.Float, UserConfigurationFieldType.String, UserConfigurationFieldType.Boolean]:
                errors["subtype"].append("Subtype can only be int, float, bool or string.")

        # Handle enums
        if choices is not None:
            if field_type != UserConfigurationFieldType.Enum:
                errors["choices"].append('The "choices" field can only be set for an "enum" type property.')
            # If the configuration parameter is of enum type, an eventual default value won't match the field type
            if default_value and default_value not in choices:
                errors["default"].append(f"{default_value} is not an available choice.")

        # Only allow the multiline property on string fields
        if "multiline" in data and field_type != UserConfigurationFieldType.String:
            errors["multiline"].append('The "multiline" field can only be set for a "string" type property.')

        # Handle everything else
        if default_value is not None and field_type != UserConfigurationFieldType.Enum:
            try:
                data_type, data_class = USER_CONFIGURATION_FIELD_TYPE_DATA_TYPE[field_type]
                if not isinstance(default_value, data_class):
                    raise ValidationError

                # In the case of model fields, the validation error is raised here if the default value is a string but not a UUID
                data_type.to_internal_value(default_value)

                # The DRF IntegerField forbids parsing a boolean as an int while the FloatField allows it
                # We prevent treating booleans as floats here to stay consistent
                if field_type == UserConfigurationFieldType.Float and isinstance(default_value, bool):
                    raise ValidationError

                # For lists, check that list elements are of given subtype
                if field_type == UserConfigurationFieldType.List and not errors.get("subtype"):
                    _, data_subclass = USER_CONFIGURATION_FIELD_TYPE_DATA_TYPE[subtype]
                    if any(not isinstance(item, data_subclass) for item in default_value):
                        errors["default"].append(f"All items in the default value must be of type {data_subclass.__name__}.")
            except (ValidationError, DjangoValidationError):
                errors["default"].append(f"This is not a valid value for a field of type {field_type.value}.")
            except KeyError:
                errors["default"].append(f"Cannot check type: {field_type.value}.")
            # Check that the UUID is that of a model that exists, for Model fields
            if field_type == UserConfigurationFieldType.Model and "default" not in errors:
                if not Model.objects.filter(id=default_value).exists():
                    errors["default"].append(f"Model {default_value} not found.")

        if errors:
            raise ValidationError(errors)
        return data


class WorkerVersionSerializer(serializers.ModelSerializer):
    """
    Serialize a worker version
    """
    # State defaults to created when instantiating a WorkerVersion
    state = EnumField(WorkerVersionState, required=False)
    worker = WorkerSerializer(read_only=True)
    revision_url = serializers.URLField(required=False, read_only=True, allow_null=True)
    gpu_usage = EnumField(FeatureUsage, required=False, default=FeatureUsage.Disabled)
    model_usage = EnumField(FeatureUsage, required=False, default=FeatureUsage.Disabled)
    # Ensure worker version configuration JSON body is an object
    configuration = serializers.DictField()

    tag = serializers.CharField(allow_null=True, max_length=512, required=False)
    branch = serializers.CharField(allow_null=True, max_length=512, required=False)

    # Serialize worker with its basic information
    class Meta:
        model = WorkerVersion
        fields = (
            "id",
            "configuration",
            "revision_url",
            "version",
            "docker_image_iid",
            "state",
            "gpu_usage",
            "model_usage",
            "worker",
            "created",
            "tag",
            "branch",
        )
        read_only_fields = ("revision_url", "version")

    def validate_tag(self, tag):
        if tag is None:
            return
        worker = self.instance.worker if self.instance else self.context["worker"]
        if worker.versions.filter(tag=tag).exists():
            raise ValidationError("A version already exists for this worker with this tag.")
        return tag

    def validate_configuration(self, configuration):
        user_configuration = configuration.get("user_configuration")
        if not user_configuration:
            return configuration

        field = serializers.DictField(child=UserConfigurationFieldSerializer())
        try:
            field.to_internal_value(user_configuration)
        except ValidationError as e:
            raise ValidationError({"user_configuration": e.detail})

        return configuration

    def validate(self, data):
        # Assert that a version set to available either has a FK to a docker image artifact or an image name
        state = data.get("state") or self.instance and self.instance.state
        if not state:
            return data

        image_iid = (
            data["docker_image_iid"] if "docker_image_iid" in data
            else self.instance and self.instance.docker_image_iid
        )
        if state == WorkerVersionState.Available and not image_iid:
            raise ValidationError({"state": [
                "A worker version must have a `docker_image_iid` set when updated to available."
            ]})

        return data

    def update(self, instance, validated_data):
        # If a branch is being set on this WorkerVersion, remove it from any version of the same worker
        if validated_data.get("branch"):
            WorkerVersion.objects.filter(worker_id=instance.worker_id, branch=validated_data["branch"]).update(branch=None)
        return super().update(instance, validated_data)


class WorkerVersionCreateSerializer(WorkerVersionSerializer):
    revision_url = serializers.URLField(
        required=False,
        allow_null=True
    )

    def validate_docker_image_iid(self, docker_image_iid):
        repository, _ = parse_repository_tag(docker_image_iid)
        try:
            resolve_repository_name(repository)
        except InvalidRepository as e:
            raise ValidationError(str(e))
        return docker_image_iid

    def validate_revision_url(self, revision_url):
        worker = self.instance.worker if self.instance else self.context["worker"]
        if worker.versions.using("default").filter(revision_url=revision_url).exists():
            raise ValidationError("A version already exists for this worker with this revision_url.")
        return revision_url

    def validate(self, data):
        super().validate(data)
        errors = defaultdict(list)

        worker = self.context["worker"]
        if worker.archived:
            errors["worker"].append("This worker is archived.")

        revision_url = data.get("revision_url")
        if isinstance(self.context["request"].auth, Task) and not revision_url:
            errors["revision_url"].append("This field is required when creating a version through Task authentication.")
        if worker.repository_url and not revision_url:
            errors["revision_url"].append("A revision url is required when creating a version for a worker linked to a repository.")

        if errors:
            raise ValidationError(errors)

        # Define a version number
        if not revision_url:
            last_version = worker.versions.using("default").aggregate(last_version=Max("version"))["last_version"]
            data["version"] = last_version + 1 if last_version else 1

        # Try to autodetect the tag from the docker image
        if "tag" not in data and "docker_image_iid" in data:
            _, tag = parse_repository_tag(data["docker_image_iid"])
            if tag is not None:
                try:
                    parse_version(tag)
                except InvalidVersion:
                    pass
                else:
                    data["tag"] = tag

        data["worker_id"] = worker.id
        return data

    def create(self, validated_data):
        # If a branch is being set on this WorkerVersion, remove it from any version of the same worker
        if validated_data.get("branch"):
            WorkerVersion.objects.filter(worker_id=validated_data["worker_id"], branch=validated_data["branch"]).update(branch=None)
        return super().create(validated_data)


class WorkerActivitySerializer(serializers.ModelSerializer):
    """
    Serialize a worker activity on an element
    """
    state = EnumField(WorkerActivityState)
    element_id = serializers.UUIDField()
    process_id = serializers.PrimaryKeyRelatedField(
        queryset=Process.objects.all(),
        # Avoid loading up to 1000 processes when opening this endpoint in a browser
        style={"base_template": "input.html"},
        required=False,
        help_text=dedent("""
            ID of the process on which the worker ran.

            Mandatory when updating an activity using WorkerVersion IDs, ignored with WorkerRun IDs.
        """),
    )
    configuration_id = serializers.UUIDField(read_only=True, allow_null=True)
    model_version_id = serializers.UUIDField(read_only=True, allow_null=True)
    worker_version_id = serializers.UUIDField(read_only=True)

    class Meta:
        model = WorkerActivity
        fields = (
            "element_id",
            "process_id",
            "configuration_id",
            "worker_version_id",
            "model_version_id",
            "created",
            "updated",
            "started",
            "state",
        )
        read_only_fields = (
            "created",
            "updated",
            "started",
            "configuration_id",
            "worker_version_id",
            "model_version_id",
        )

    def validate_process_id(self, process):
        request = self.context.get("request")
        if request and isinstance(request.auth, Task) and process.id != request.auth.process_id:
            raise serializers.ValidationError("Only WorkerActivities for the process of the currently authenticated task can be updated.")
        return process


class WorkerConfigurationListSerializer(serializers.ModelSerializer):
    configuration = serializers.DictField(allow_empty=False)

    class Meta:
        model = WorkerConfiguration
        fields = (
            "id",
            "name",
            "configuration",
            "archived"
        )

    def validate(self, data):
        # Check unique constraints
        name = data.get("name")
        configuration = data.get("configuration")
        configuration_hash = hash_object(configuration)

        # No further validation required if we only sent `archived`
        if not name and not configuration:
            return data

        if "worker" in self.context:
            # CreateWorkerConfiguration provides the worker through the context
            worker = self.context["worker"]
            data["worker_id"] = worker.id
        elif self.instance is not None:
            # Use the already set worker in (Partial)UpdateConfiguration to revalidate
            worker = self.instance.worker

        existing_configurations = list(worker.configurations.filter(Q(configuration_hash=configuration_hash) | Q(name=name)))
        errors = defaultdict(list)
        for config in existing_configurations:
            if configuration_hash == config.configuration_hash:
                errors["configuration"].append("A worker configuration with this configuration already exists for this worker.")
                errors["id"] = config.id
            if name == config.name:
                errors["name"].append("A worker configuration with this name already exists for this worker.")

        # Due to the defaultdict, writing e.g. len(errors['configuration']) > 1 would add {'configuration': []} to
        # the errors dict if the key is not already in the dict.
        if len(errors.get("name", [])) > 1 or len(errors.get("configuration", [])) > 1:
            # If there is more than one configuration with the same name and/or configuration as the one the user
            # is trying to create, raise a MultipleObjectsReturned exception because this shouldn't be possible
            # due to the unicity constraints. There can however be more than one worker configurations in
            # existing_configurations, if both a worker configuration with the same name and the same configuration
            # exist; but checking the length of existing_configurations wouldn't differentiate between this case
            # and an actual duplicated worker configuration.
            raise WorkerConfiguration.MultipleObjectsReturned

        # Archived configurations cannot be renamed, but un-archiving a configuration and renaming it at the same time is
        # possible. It's also possible to archive and rename a configuration simultaneously.
        if self.instance is not None and self.instance.archived and data.get("archived") is not False and name:
            errors["name"].append("Archived configurations cannot be renamed.")

        if errors:
            raise ValidationError(errors)

        return data


class WorkerConfigurationSerializer(WorkerConfigurationListSerializer):
    configuration = serializers.DictField(allow_empty=False, read_only=True)

    class Meta(WorkerConfigurationListSerializer.Meta):
        # The configuration cannot be updated
        read_only_fields = ("id", "configuration")


class CorpusWorkerVersionSerializer(serializers.ModelSerializer):
    worker_version = WorkerVersionSerializer()
    model_version = ModelVersionLightSerializer(read_only=True, allow_null=True)
    worker_configuration = WorkerConfigurationSerializer(read_only=True, allow_null=True)

    class Meta:
        model = CorpusWorkerVersion
        fields = ("id", "worker_version", "model_version", "worker_configuration")


class WorkerConfigurationExistsErrorSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="UUID of an existing worker configuration, if the error comes from a duplicate configuration.")
    configuration = serializers.ListField(child=serializers.CharField(), required=False, help_text="Configuration error message.")
    name = serializers.ListField(child=serializers.CharField(), required=False, help_text="Name error message.")


class WorkerStatisticsSerializer(serializers.Serializer):
    """
    Serialize activity statistics of a worker version
    """
    worker_version_id = serializers.UUIDField(read_only=True)
    configuration_id = serializers.UUIDField(read_only=True, allow_null=True)
    model_version_id = serializers.UUIDField(read_only=True, allow_null=True)
    queued = serializers.IntegerField(read_only=True)
    started = serializers.IntegerField(read_only=True)
    processed = serializers.IntegerField(read_only=True)
    error = serializers.IntegerField(read_only=True)
    average_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        # Also document the timestamp format, specified by a random Django util function used by DRF:
        # https://github.com/django/django/blob/eccda63a49b92ab14d523d69da2a3b3246cf91d1/django/utils/duration.py#L18
        help_text=dedent("""
        Average time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        """),
    )
    min_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent("""
        Shortest time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        """),
    )
    max_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent("""
        Longest time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        """),
    )
    median_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent("""
        Median time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        """),
    )
    completion_estimate_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent("""
        Estimated time for remaining elements to be processed based on median process time.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        """),
    )
